#!/usr/bin/python
#
# Copyright (C) 2009 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines event types that are sent from the wave server.

This module defines all of the event types currently supported by the wave
server. Each event type is sub classed from Event and has its own
properties depending on the type.
"""


class Context(object):
  """Specifies constants representing different context requests."""

  #: Requests the root blip.
  ROOT = 'ROOT'
  #: Requests the parent blip of the event blip.
  PARENT = 'PARENT'
  #: Requests the siblings blip of the event blip.
  SIBLINGS = 'SIBLINGS'
  #: Requests the child blips of the event blip.
  CHILDREN = 'CHILDREN'
  #: Requests the event blip itself.
  SELF = 'SELF'
  #: Requests all of the blips of the event wavelet.
  ALL = 'ALL'


class Event(object):
  """Object describing a single event.

  Attributes:
    modified_by: Participant id that caused this event.

    timestamp: Timestamp that this event occurred on the server.

    type: Type string of this event.

    properties: Dictionary of all extra properties. Typically the derrived
        event type should have these explicitly set as attributes, but
        experimental features might appear in properties before that.

    blip_id: The blip_id of the blip for blip related events or the root
        blip for wavelet related events.

    blip: If available, the blip with id equal to the events blip_id.

    proxying_for: If available, the proxyingFor id of the robot that caused the
    event.
  """

  def __init__(self, json, wavelet):
    """Inits this event with JSON data.

    Args:
      json: JSON data from Wave server.
    """
    self.modified_by = json.get('modifiedBy')
    self.timestamp = json.get('timestamp', 0)
    self.type = json.get('type')
    self.raw_data = json
    self.properties = json.get('properties', {})
    self.blip_id = self.properties.get('blipId')
    self.blip = wavelet.blips.get(self.blip_id)
    self.proxying_for = json.get('proxyingFor')

class WaveletBlipCreated(Event):
  """Event triggered when a new blip is created.

  Attributes:
    new_blip_id: The id of the newly created blip.

    new_blip: If in context, the actual new blip.
  """
  type = 'WAVELET_BLIP_CREATED'

  def __init__(self, json, wavelet):
    super(WaveletBlipCreated, self).__init__(json, wavelet)
    self.new_blip_id = self.properties['newBlipId']
    self.new_blip = wavelet.blips.get(self.new_blip_id)


class WaveletBlipRemoved(Event):
  """Event triggered when a new blip is removed.

  Attributes:
    removed_blip_id: the id of the removed blip

    removed_blip: if in context, the removed blip
  """
  type = 'WAVELET_BLIP_REMOVED'

  def __init__(self, json, wavelet):
    super(WaveletBlipRemoved, self).__init__(json, wavelet)
    self.removed_blip_id = self.properties['removedBlipId']
    self.removed_blip = wavelet.blips.get(self.removed_blip_id)


class WaveletParticipantsChanged(Event):
  """Event triggered when the participants on a wave change.

  Attributes:
    participants_added: List of participants added.

    participants_removed: List of participants removed.
  """
  type = 'WAVELET_PARTICIPANTS_CHANGED'

  def __init__(self, json, wavelet):
    super(WaveletParticipantsChanged, self).__init__(json, wavelet)
    self.participants_added = self.properties['participantsAdded']
    self.participants_removed = self.properties['participantsRemoved']


class WaveletSelfAdded(Event):
  """Event triggered when the robot is added to the wavelet."""
  type = 'WAVELET_SELF_ADDED'


class WaveletSelfRemoved(Event):
  """Event triggered when the robot is removed from the wavelet."""
  type = 'WAVELET_SELF_REMOVED'


class WaveletTitleChanged(Event):
  """Event triggered when the title of the wavelet has changed.

  Attributes:
    title: The new title.
  """
  type = 'WAVELET_TITLE_CHANGED'

  def __init__(self, json, wavelet):
    super(WaveletTitleChanged, self).__init__(json, wavelet)
    self.title = self.properties['title']


class BlipContributorsChanged(Event):
  """Event triggered when the contributors to this blip change.

  Attributes:
    contributors_added: List of contributors that were added.

    contributors_removed: List of contributors that were removed.
  """
  type = 'BLIP_CONTRIBUTORS_CHANGED'

  def __init__(self, json, wavelet):
    super(BlipContributorsChanged, self).__init__(json, wavelet)
    self.contibutors_added = self.properties['contributorsAdded']
    self.contibutors_removed = self.properties['contributorsRemoved']


class BlipSubmitted(Event):
  """Event triggered when a blip is submitted."""
  type = 'BLIP_SUBMITTED'


class DocumentChanged(Event):
  """Event triggered when a document is changed.

  This event is fired after any changes in the document and should be used
  carefully to keep the amount of traffic to the robot reasonable. Use
  filters where appropriate.
  """
  type = 'DOCUMENT_CHANGED'


class FormButtonClicked(Event):
  """Event triggered when a form button is clicked.

  Attributes:
    button_name: The name of the button that was clicked.
  """
  type = 'FORM_BUTTON_CLICKED'

  def __init__(self, json, wavelet):
    super(FormButtonClicked, self).__init__(json, wavelet)
    self.button_name = self.properties['buttonName']


class GadgetStateChanged(Event):
  """Event triggered when the state of a gadget changes.

  Attributes:
    index: The index of the gadget that changed in the document.

    old_state: The old state of the gadget.
  """
  type = 'GADGET_STATE_CHANGED'

  def __init__(self, json, wavelet):
    super(GadgetStateChanged, self).__init__(json, wavelet)
    self.index = self.properties['index']
    self.old_state = self.properties['oldState']


class AnnotatedTextChanged(Event):
  """Event triggered when text with an annotation has changed.

  This is mainly useful in combination with a filter on the
  name of the annotation.

  Attributes:
    name: The name of the annotation.

    value: The value of the annotation that changed.
  """
  type = 'ANNOTATED_TEXT_CHANGED'

  def __init__(self, json, wavelet):
    super(AnnotatedTextChanged, self).__init__(json, wavelet)
    self.name = self.properties['name']
    self.value = self.properties.get('value')


class OperationError(Event):
  """Triggered when an event on the server occurred.

  Attributes:
    operation_id: The operation id of the failing operation.

    error_message: More information as to what went wrong.
  """
  type = 'OPERATION_ERROR'

  def __init__(self, json, wavelet):
    super(OperationError, self).__init__(json, wavelet)
    self.operation_id = self.properties['operationId']
    self.error_message = self.properties['message']


class WaveletCreated(Event):
  """Triggered when a new wavelet is created.

  This event is only triggered if the robot creates a new
  wavelet and can be used to initialize the newly created wave.
  wavelets created by other participants remain invisible
  to the robot until the robot is added to the wave in
  which case WaveletSelfAdded is triggered.

  Attributes:
    message: Whatever string was passed into the new_wave
        call as message (if any).
  """
  type = 'WAVELET_CREATED'

  def __init__(self, json, wavelet):
    super(WaveletCreated, self).__init__(json, wavelet)
    self.message = self.properties['message']


class WaveletFetched(Event):
  """Triggered when a new wavelet is fetched.

  This event is triggered after a robot requests to
  see another wavelet. The robot has to be on the other
  wavelet already.

  Attributes:
    message: Whatever string was passed into the new_wave
        call as message (if any).
  """
  type = 'WAVELET_FETCHED'

  def __init__(self, json, wavelet):
    super(WaveletFetched, self).__init__(json, wavelet)
    self.message = self.properties['message']


class WaveletTagsChanged(Event):
  """Event triggered when the tags on a wavelet change."""
  type = 'WAVELET_TAGS_CHANGED'

  def __init__(self, json, wavelet):
    super(WaveletTagsChanged, self).__init__(json, wavelet)


def is_event(cls):
  """Returns whether the passed class is an event."""
  try:
    if not issubclass(cls, Event):
      return False
    return hasattr(cls, 'type')
  except TypeError:
    return False

ALL = [item for item in globals().copy().values() if is_event(item)]



# me2/plugs/plug.py
#
#

""" plugin management. """

## me2 imports

from me2.lib.commands import cmnds
from me2.lib.examples import examples
from me2.lib.boot import default_plugins, plugin_packages, remove_plugin, update_mod

## plug-enable command

def handle_plugenable(bot, event):
    if not event.rest: event.missing("<plugin>") ; return
    mod = bot.plugs.getmodule(event.rest)
    if not mod: event.reply("can't find module for %s" % event.rest) ; return
    event.reply("reloading and enabling %s" % mod)
    bot.enable(mod)
    bot.plugs.load(mod, force=True)
    #update_mod(mod)
    event.done()

cmnds.add("plug-enable", handle_plugenable, ["OPER", ])
examples.add("plug-enable", "enable a plugin", "plug-enable me2.plugs.common.rss")

## plug-disable command

def handle_plugdisable(bot, event):
    if not event.rest: event.missing("<plugin>") ; return
    mod = bot.plugs.getmodule(event.rest)
    if mod in default_plugins: event.reply("can't remove a default plugin") ; return
    if not mod: event.reply("can't find module for %s" % event.rest) ; return
    event.reply("unloading and disabling %s" % mod)
    bot.plugs.unload(mod)
    bot.disable(mod)
    #remove_plugin(mod)
    event.done()

cmnds.add("plug-disable", handle_plugdisable, ["OPER", ])
examples.add("plug-disable", "disable a plugin", "plug-disable me2.plugs.common.rss")



# me2/fleet.py
#
#

""" fleet is a list of bots. """

## me2 imports

from me2.utils.exception import handle_exception
from me2.utils.generic import waitforqueue
from config import Config
from users import users
from plugins import plugs
from persist import Persist
from errors import NoSuchBotType, BotNotEnabled
from threads import start_new_thread
from eventhandler import mainhandler
from me2.utils.name import stripname

## simplejson imports

from me2.contrib.simplejson import load

## basic imports

import Queue
import os
import types
import time
import glob
import logging
import threading

## classes

class FleetBotAlreadyExists(Exception):
    pass

class Fleet(Persist):

    """
        a fleet contains multiple bots (list of bots).

    """

    def __init__(self, datadir):
        Persist.__init__(self, datadir + os.sep + 'fleet' + os.sep + 'fleet.main')
        if not self.data.has_key('names'): self.data['names'] = []
        if not self.data.has_key('types'): self.data['types'] = {}
        self.startok = threading.Event()
        self.bots = []

    def loadall(self, names=[]):
        """ load all bots. """ 
        target = names or self.data.names
        if not target: logging.error("fleet - no bots in fleet")
        else: logging.warning("fleet - loading %s" % ", ".join(target))
        for name in target:
            if not name: logging.debug("fleet - name is not set") ; continue
            try:
                if self.data.types[name] == "console": logging.warn("fleet- skipping console bot %s" % name) ; continue
            except KeyError: continue
            try: self.makebot(self.data.types[name], name)
            except BotNotEnabled: pass
            except KeyError: logging.error("no type know for %s bot" % name)
            except Exception, ex: handle_exception()

    def avail(self):
        """ return available bots. """
        return self.data['names']

    def getfirstbot(self):
        """ return the first bot in the fleet. """
        return self.bots[0]

    def getfirstjabber(self, isgae=False):
        """ return the first jabber bot of the fleet. """
        for bot in self.bots:
            if isgae:
                if (bot.type == 'xmpp' or bot.type == 'jabber'):  return bot
            elif bot.type == 'sxmpp': return bot
        
    def size(self):
        """ return number of bots in fleet. """
        return len(self.bots)

    def settype(self, name, type):
        """ set the type of a bot. """
        cfg = Config('fleet' + os.sep + stripname(name) + os.sep + 'config')
        cfg['name'] = name
        logging.debug("fleet - %s - setting type to %s" % (self.cfile, type))
        cfg.type = type
        cfg.save()

    def makebot(self, type, name, domain="", cfg={}):
        """ create a bot .. use configuration if provided. """
        if cfg: logging.warn('fleet - making %s (%s) bot - %s' % (type, name, cfg.dump()))
        bot = None
        if not cfg:
            cfg = Config('fleet' + os.sep + stripname(name) + os.sep + 'config')
            cfg['name'] = cfg['botname'] = name
        if not cfg.type and type:
            logging.debug("fleet - %s - setting type to %s" % (cfg.cfile, type))
            cfg.type = type
        if not cfg['type']:
            try:
                self.data['names'].remove(name)
                self.save()
            except ValueError: pass
            raise Exception("no bot type specified")
        if not cfg.owner:
            cfg.owner = Config().owner
        if not cfg.domain and domain: cfg.domain = domain
        if not cfg: raise Exception("can't make config for %s" % name)
        cfg.save()
        if cfg.disable:
            logging.warn("fleet - %s bot is disabled" % name)
            raise BotNotEnabled(name)
        if type == 'xmpp' or type == 'jabber':
            try:
                from me2.lib.gae.xmpp.bot import XMPPBot
                bot = XMPPBot(cfg)
            except ImportError:
                from me2.lib.socklib.xmpp.bot import SXMPPBot          
                bot = SXMPPBot(cfg)
        elif type == 'sxmpp':
            from me2.lib.socklib.xmpp.bot import SXMPPBot
            bot = SXMPPBot(cfg)
        elif type == 'web':
            from me2.lib.gae.web.bot import WebBot
            bot = WebBot(cfg)
        elif type == 'wave':
            from me2.lib.gae.wave.bot import WaveBot
            dom = cfg.domain or domain
            bot = WaveBot(cfg, domain=dom)
        elif type == 'irc':
            from me2.lib.socklib.irc.bot import IRCBot
            bot = IRCBot(cfg)
        elif type == 'console':
            from me2.lib.console.bot import ConsoleBot
            bot = ConsoleBot(cfg)
        elif type == 'base':
            from me2.lib.botbase import BotBase
            bot = BotBase(cfg)
        else: raise NoSuchBotType('%s bot .. unproper type %s' % (name, type))
        if bot:
            self.addbot(bot)
            return bot
        raise Exception("can't make %s bot" % name)

    def save(self):
        """ save fleet data and call save on all the bots. """
        Persist.save(self)
        for i in self.bots:
            try: i.save()
            except Exception, ex: handle_exception()

    def list(self):
        """ return list of bot names. """
        result = []
        for i in self.bots: result.append(i.name)
        return result

    def stopall(self):
        """ call stop() on all fleet bots. """
        for i in self.bots:
            try: i.stop()
            except: handle_exception()

    def byname(self, name):
        """ return bot by name. """
        for i in self.bots:
            if name == i.name: return i

    def replace(self, name, bot):
        """ replace bot with a new bot. """
        for i in range(len(self.bots)):
            if name == self.bots[i].name:
                self.bots[i] = bot
                return True

    def enable(self, cfg):
        """ enable a bot baed of provided config. """
        if cfg.botname and cfg.botname not in self.data['names']:
            self.data['names'].append(cfg.botname)
            self.data['types'][cfg.botname] = cfg.type
            self.save()
        return True

    def addbot(self, bot):
        """
            add a bot to the fleet .. remove all existing bots with the 
            same name.

        """
        for i in range(len(self.bots)-1, -1, -1):
            if self.bots[i].name == bot.botname:
                logging.debug('fleet - removing %s from fleet' % bot.botname)
                del self.bots[i]
        logging.info('fleet - adding %s' % bot.botname)
        self.bots.append(bot)
        if bot.botname and bot.botname not in self.data['names']:
            self.data['names'].append(bot.botname)
            self.data['types'][bot.botname] = bot.type
            self.save()
        return True

    def delete(self, name):
        """ delete bot with name from fleet. """
        for bot in self.bots:
            if bot.botname == name:
                bot.exit()
                self.remove(i)
                bot.cfg['disable'] = 1
                bot.cfg.save()
                logging.debug('fleet - %s disabled' % bot.botname)
                return True
        return False

    def remove(self, bot):
        """ delete bot by object. """
        try:
            self.bots.remove(bot)
            return True
        except ValueError:
            return False

    def exit(self, name=None, jabber=False):
        """ call exit on all bots. """
        if not name:
            threads = []
            for bot in self.bots:
                if jabber and bot.type != 'sxmpp' and bot.type != 'jabber': continue
                threads.append(start_new_thread(bot.exit, ()))
            for thread in threads: thread.join()
            return
        for bot in self.bots:
            if bot.botname == name:
                if jabber and bot.type != 'sxmpp' and bot.type != 'jabber': continue
                try: bot.exit()
                except: handle_exception()
                self.remove(bot)
                return True
        return False

    def cmnd(self, event, name, cmnd):
        """ do command on a bot. """
        bot = self.byname(name)
        if not bot: return 0
        from me2.lib.eventbase import EventBase
        j = plugs.clonedevent(bot, event)
        j.onlyqueues = True
        j.txt = cmnd
        q = Queue.Queue()
        j.queues = [q]
        j.speed = 3
        plugs.trydispatch(bot, j)
        result = waitforqueue(q)
        if not result: return
        res = ["[%s]" % bot.botname, ]
        res += result
        event.reply(res)
        return res

    def cmndall(self, event, cmnd):
        """ do a command on all bots. """
        for bot in self.bots: self.cmnd(event, bot.botname, cmnd)

    def broadcast(self, txt):
        """ broadcast txt to all bots. """
        for bot in self.bots: bot.broadcast(txt)

    def startall(self):
        for bot in self.bots: start_new_thread(bot.start, ())

    def resume(self, sessionfile):
        """ resume bot from session file. """
        session = load(open(sessionfile))
        for name in session['bots'].keys():
            reto = None
            if session['name'] == name: reto = session['channel']
            start_new_thread(self.resumebot, (name, session['bots'][name], reto))
        time.sleep(5)
        self.startok.set()

    def resumebot(self, botname, data={}, printto=None):
        """
            resume single bot. """
        logging.warn("fleet - resuming %s bot" % botname)
        # see if we need to exit the old bot
        if data['type'] == "console": logging.warn("not resuming console bot %s" % botname) ; return
        oldbot = self.byname(botname)
        if oldbot and data['type'] == "sxmpp": oldbot.exit()
        cfg = Config('fleet' + os.sep + stripname(botname) + os.sep + 'config')
        logging.warn("fleet - resuming %s bot - %s - %s" % (botname, str(data), data['type']))
        bot = self.makebot(data['type'], botname)
        if data['type'] != "sxmpp":
            if oldbot: self.replace(oldbot, bot)
            bot._resume(data, printto)
            bot.start(False)
        else: start_new_thread(bot.start, ())

## global fleet object

fleet = None

def getfleet(datadir=None, new=False):
    if not datadir:
         from me2.lib.datadir import getdatadir 
         datadir = getdatadir()
    global fleet
    if not fleet or new: fleet = Fleet(datadir)
    return fleet



#!/usr/bin/python2.4
#
# Copyright (C) 2009 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility library containing various helpers used by the API."""

import re

CUSTOM_SERIALIZE_METHOD_NAME = 'serialize'

MARKUP_RE = re.compile(r'<([^>]*?)>')


def force_unicode(object):
  """ Return the Unicode string version of object, with UTF-8 encoding. """
  if isinstance(object, unicode):
    return object
  return unicode(str(object), 'utf-8')

def parse_markup(markup):
  """Parses a bit of markup into robot compatible text.
  
  For now this is a rough approximation.
  """
  def replace_tag(group):
    if not group.groups:
      return ''
    tag = group.groups()[0].split(' ', 1)[0]
    if (tag == 'p' or tag == 'br'):
      return '\n'
    return ''

  return MARKUP_RE.sub(replace_tag, markup)

def is_iterable(inst):
  """Returns whether or not this is a list, tuple, set or dict .

  Note that this does not return true for strings.
  """
  return hasattr(inst, '__iter__')

def is_dict(inst):
  """Returns whether or not the specified instance is a dict."""
  return hasattr(inst, 'iteritems')


def is_user_defined_new_style_class(obj):
  """Returns whether or not the specified instance is a user-defined type."""
  return type(obj).__module__ != '__builtin__'

def lower_camel_case(s):
  """Converts a string to lower camel case.

  Examples:
    foo => foo
    foo_bar => fooBar
    foo__bar => fooBar
    foo_bar_baz => fooBarBaz

  Args:
    s: The string to convert to lower camel case.

  Returns:
    The lower camel cased string.
  """
  return reduce(lambda a, b: a + (a and b.capitalize() or b), s.split('_'))

def non_none_dict(d):
  """return a copy of the dictionary without none values."""
  return dict([a for a in d.items() if not a[1] is None])

def _serialize_attributes(obj):
  """Serializes attributes of an instance.

  Iterates all attributes of an object and invokes serialize if they are
  public and not callable.

  Args:
    obj: The instance to serialize.

  Returns:
    The serialized object.
  """
  data = {}
  for attr_name in dir(obj):
    if attr_name.startswith('_'):
      continue
    attr = getattr(obj, attr_name)
    if attr is None or callable(attr):
      continue
    # Looks okay, serialize it.
    data[lower_camel_case(attr_name)] = serialize(attr)
  return data


def _serialize_dict(d):
  """Invokes serialize on all of its key/value pairs.

  Args:
    d: The dict instance to serialize.

  Returns:
    The serialized dict.
  """
  data = {}
  for k, v in d.items():
    data[lower_camel_case(k)] = serialize(v)
  return data


def serialize(obj):
  """Serializes any instance.

  If this is a user-defined instance
  type, it will first check for a custom Serialize() function and use that
  if it exists. Otherwise, it will invoke serialize all of its public
  attributes. Lists and dicts are serialized trivially.

  Args:
    obj: The instance to serialize.

  Returns:
    The serialized object.
  """
  if is_user_defined_new_style_class(obj):
    if obj and hasattr(obj, CUSTOM_SERIALIZE_METHOD_NAME):
      method = getattr(obj, CUSTOM_SERIALIZE_METHOD_NAME)
      if callable(method):
        return method()
    return _serialize_attributes(obj)
  elif is_dict(obj):
    return _serialize_dict(obj)
  elif is_iterable(obj):
    return [serialize(v) for v in obj]
  return obj


class StringEnum(object):
  """Enum like class that is configured with a list of values.

  This class effectively implements an enum for Elements, except for that
  the actual values of the enums will be the string values.
  """

  def __init__(self, *values):
    for name in values:
      setattr(self, name, name)



# handler_task.py
#
#

""" me2 task handler. """

## me2 imports

from me2.lib.plugins import plugs
from me2.lib.version import getversion
from me2.utils.exception import handle_exception
from me2.lib.tasks import taskmanager

## google imports

from google.appengine.ext import webapp
from google.appengine.ext.webapp.util import run_wsgi_app

## simplejson import

from waveapi.simplejson import loads

## basic imports

import wsgiref.handlers
import logging
import google

## vars

periodicals =  ['me2.plugs.common.rss', 'me2.plugs.botevent']
mountpoints = ['rss', 'botevent']

##

logging.info(getversion('TASK'))

for plugin in periodicals:
    plugs.reload(plugin)


class TaskHandler(webapp.RequestHandler):

    """ the bots task handler. """

    def get(self):
        try:
            """ this is where the task gets dispatched. """
            path = self.request.path
            if path.endswith('/'):
                path = path[:-1]
            taskname = path.split('/')[-1].strip()
            logging.debug("using taskname: %s" % taskname)

            inputdict = {}
            for name, value in self.request.environ.iteritems():
                if not 'wsgi' in name:
                    inputdict[name] = value

            taskmanager.dispatch(taskname, inputdict, self.request, self.response)

        except google.appengine.runtime.DeadlineExceededError:
            return
        except Exception, ex:
            handle_exception()
            #self.send_error(500)

    def post(self):
        """ this is where the task gets dispatched. """
        try:
            path = self.request.path
            if path.endswith('/'):
                path = path[:-1]
            taskname = path.split('/')[-1].strip()
            logging.debug("using taskname: %s taken from %s" % (taskname, path))
            if not taskname:
                return

            inputdict = {}
            for name, value in self.request.environ.iteritems():
                if not 'wsgi' in name:
                    inputdict[name] = value

            taskmanager.dispatch(taskname, inputdict, self.request, self.response)
        except google.appengine.runtime.DeadlineExceededError:
            return
        except Exception, ex:
            handle_exception()
            #self.send_error(500)

# the application 

mountlist = []

for mount in mountpoints:
    mountlist.append(('/tasks/%s' % mount, TaskHandler))

application = webapp.WSGIApplication(mountlist, debug=True)

def main():
    global application
    run_wsgi_app(application)

if __name__ == "__main__":
    main()



# me2/plugs/core.py
#
#

""" core bot commands. """

## gozerbot imports

from me2.utils.statdict import StatDict
from me2.utils.log import setloglevel
from me2.utils.timeutils import elapsedstring
from me2.utils.exception import handle_exception
from me2.lib.commands import cmnds
from me2.lib.examples import examples
from me2.lib.plugins import plugs
from me2.lib.boot import plugin_packages, getpluginlist, boot, getcmndtable, whatcommands
from me2.lib.persist import Persist
from me2.lib.reboot import reboot, reboot_stateful
from me2.lib.eventhandler import mainhandler
from me2.lib.fleet import getfleet
from me2.lib.socklib.partyline import partyline
from me2.lib.exit import globalshutdown
from me2.lib.runner import defaultrunner, cmndrunner, longrunner
from me2.lib.errors import NoSuchPlugin

## basic imports

import time
import threading
import sys
import re
import os
import copy
import cgi

## define

cpy = copy.deepcopy

## reboot command

def handle_reboot(bot, ievent):
    """ reboot the bot. """
    if bot.isgae:
        ievent.reply("this command doesn't work on the GAE")
        return
    ievent.reply("rebooting")
    time.sleep(3)
    if ievent.rest == "cold": stateful = False
    else: stateful = True
    if stateful: mainhandler.put(0, reboot_stateful, bot, ievent, getfleet(), partyline)
    else:
        bot.exit()
        mainhandler.put(0, reboot)

cmnds.add("reboot", handle_reboot, "OPER")
examples.add("reboot", "reboot the bot.", "reboot")

## quit command

def handle_quit(bot, ievent):
    """ disconnect from the server. """
    if bot.isgae:
        ievent.reply("this command doesnt work on the GAE")
        return
    ievent.reply("quiting")
    bot.exit()

cmnds.add("quit", handle_quit, "OPER")
examples.add("quit", "quit the bot.", "quit")

## encoding command

def handle_encoding(bot, ievent):
    """ show default encoding. """
    ievent.reply('default encoding is %s' % bot.encoding or sys.getdefaultencoding())

cmnds.add('encoding', handle_encoding, ['USER', 'GUEST'])
examples.add('encoding', 'show default encoding', 'encoding')

## uptime command

def handle_uptime(bot, ievent):
    """ show uptime. """
    ievent.reply("<b>uptime is %s<b>" % elapsedstring(time.time()-bot.starttime))

cmnds.add('uptime', handle_uptime, ['USER', 'GUEST'])
examples.add('uptime', 'show uptime of the bot', 'uptime')

## list command

def handle_available(bot, ievent):
    """ show available plugins .. to enable use !reload. """
    if ievent.rest: ievent.reply("%s plugin has the following commands: " % ievent.rest, whatcommands(ievent.rest))
    else: ievent.reply("available plugins: ", getpluginlist(), raw=True) ; return

cmnds.add('list', handle_available, ['USER', 'GUEST'])
examples.add('list', 'list available plugins', 'list')

## commands command

def handle_commands(bot, ievent):
    """ show commands of plugin. """
    try: plugin = ievent.args[0].lower()
    except IndexError: plugin = ""
    result = []
    cmnds = getcmndtable()
    for cmnd, plugname in cmnds.iteritems(): 
        if plugname:
            if not plugin or plugin in plugname: result.append(cmnd)
    if result:
        result.sort()
        if not plugin: plugin = "BOTME2"
        ievent.reply('%s has the following commands: ' % plugin, result)
    else: ievent.reply('no commands found for plugin %s' % plugin)

cmnds.add('commands', handle_commands, ['USER', 'GUEST'])
examples.add('commands', 'show commands of <plugin>', '1) commands core')

## perm command

def handle_perm(bot, ievent):
    """ get permission of command. """
    try:cmnd = ievent.args[0]
    except IndexError:
        ievent.missing("<cmnd>")
        return
    try: perms = cmnds.perms(cmnd)
    except KeyError:
        ievent.reply("no %sw command registered")
        return
    if perms: ievent.reply("%s command needs %s permission" % (cmnd, perms))
    else: ievent.reply("can't find perm for %s" % cmnd)

cmnds.add('perm', handle_perm, ['USER', 'GUEST'])
examples.add('perm', 'show permission of command', 'perm quit')

## version command

def handle_version(bot, ievent):
    """ show bot's version. """
    from me2.lib.version import getversion
    ievent.reply(getversion(bot.type.upper()))

cmnds.add('version', handle_version, ['USER', 'GUEST'])
examples.add('version', 'show version of the bot', 'version')

## whereis command

def handle_whereis(bot, ievent):
    """ locate a command. """
    try: cmnd = ievent.args[0]
    except IndexError:
        ievent.missing('<cmnd>')
        return
    plugin = cmnds.whereis(cmnd)
    if plugin: ievent.reply("%s command is in: %s" %  (cmnd, plugin))
    else: ievent.reply("can't find " + cmnd)

cmnds.add('whereis', handle_whereis, ['USER', 'GUEST'])
examples.add('whereis', 'whereis <cmnd> .. show in which plugins <what> is', 'whereis test')

## help command

def handle_helpplug(bot, ievent):
    """ how help on plugin/command or show basic help msg. """
    try: what = ievent.args[0]
    except (IndexError, TypeError):
        ievent.reply("available plugins: ", getpluginlist())
        ievent.reply("see !help <plugin> to get help on a plugin.")
        return
    plugin = None
    modname = ""
    perms = []
    for package in plugin_packages:
        try:
             modname = "%s.%s" % (package, what)
             try:
                 plugin = plugs.reload(modname)
                 if plugin: break
             except NoSuchPlugin: continue
        except(KeyError, ImportError): pass
    if not plugin:
        ievent.reply("no %s plugin loaded" % what)
        return
    try: phelp = plugin.__doc__
    except (KeyError, AttributeError):
        ievent.reply('no description of %s plugin available' % what)
        return
    cmndresult = []
    if phelp:
        perms = ievent.user.data.perms
        if not perms: perms = ['GUEST', ]
        counter = 1
        for i, j in cmnds.iteritems():
            if what == j.plugname:
                for perm in j.perms:
                    if perm in perms:
                        if True:
                            try:
                                descr = j.func.__doc__
                                if not descr: descr = "no description provided"
                                try: cmndresult.append(u"    <b>!%s</b> - <i>%s</i> - examples: %s" % (i, descr, examples[i].example))
                                except KeyError: cmndresult.append(u"    <b>!%s</b> - <i>%s</i> - no examples" % (i, descr))
                            except AttributeError: pass
                            counter += 1
                            break
    if cmndresult and phelp:
        res = []
        for r in cmndresult:
            if bot.type in ['web', ]: res.append("%s<br>" % r)
            else: res.append(r)
        res.sort()
        what = what.upper()
        res.insert(0, "%s\n" % phelp.strip())
        ievent.reply('HELP ON %s \n\n' % what, res, dot="\n")
    else:
        if perms: ievent.reply('no commands available for permissions: %s' % ", ".join(perms))
        else: ievent.reply("can't find help on %s" % what)

cmnds.add('help', handle_helpplug, ['USER', 'GUEST'])
examples.add('help', 'get help on <cmnd> or <plugin>', '1) help 2) help misc')

## apro command

def handle_apro(bot, ievent):
    """ apropos for command. """
    try: what = ievent.args[0]
    except IndexError:
        ievent.missing('<what>')
        return
    result = []
    cmnds = getcmndtable()
    for i in cmnds.keys():
        if what in i: result.append(i)
    result.sort()
    if result: ievent.reply("commands matching %s: " % what, result)
    else: ievent.reply('no matching commands found for %s' % what)

cmnds.add('apro', handle_apro, ['USER', 'GUEST'])
examples.add('apro', 'apro <what> .. search for commands that contain <what>', 'apro com')

## whatcommands command

def handle_whatcommands(bot, ievent):
    """ show all commands with permission. """
    if not ievent.rest:
        ievent.missing('<perm>')
        return
    result = cmnds
    res = []
    for cmnd in result.values():
        if cmnd and cmnd.perms and ievent.rest in cmnd.perms:
            res.append(cmnd.cmnd)
    res.sort()
    if not res: ievent.reply('no commands known for permission %s' % ievent.rest)
    else: ievent.reply('commands known for permission %s: ' % ievent.rest, res)

cmnds.add('whatcommands', handle_whatcommands, ['USER', 'GUEST'])
examples.add('whatcommands', 'show commands with permission <perm>', 'whatcommands USER')

## versions command

def handle_versions(bot, ievent):
    """ show versions of all loaded modules (if available). """
    versions = {}
    for mod in copy.copy(sys.modules):
        try: versions[mod] = sys.modules[mod].__version__
        except AttributeError, ex: pass
        try: versions['python'] = sys.version
        except AttributeError, ex: pass
    ievent.reply("versions ==> %s" % unicode(versions))

cmnds.add('versions', handle_versions, ['USER', 'GUEST'])
examples.add('versions', 'show versions of all loaded modules', 'versions')

## loglevel command

def handle_loglevel(bot, event):
    """ change loglevel of the bot. """
    if not event.rest:
        event.missing("<loglevel> (string)")
        return
    setloglevel(event.rest)
    event.done()

cmnds.add("loglevel", handle_loglevel, "OPER")
examples.add("loglevel", "set loglevel ot on of debug, info, warning or error", "loglevel warn")

## threads command

def handle_threads(bot, ievent):
    """ show running threads. """
    try: import threading
    except ImportError:
         ievent.reply("threading is not enabled.")
         return
    stats = StatDict()
    threadlist = threading.enumerate()
    for thread in threadlist: stats.upitem(thread.getName())
    result = []
    for item in stats.top(): result.append("%s = %s" % (item[0], item[1]))
    result.sort()
    ievent.reply("threads running: ", result)

cmnds.add('threads', handle_threads, ['USER', 'GUEST'])
examples.add('threads', 'show running threads', 'threads')

## loaded command

def handle_loaded(bot, event):
    """ show plugins in cache. """
    event.reply("loaded plugins (cache): ", plugs.keys())

cmnds.add('loaded', handle_loaded, ['USER', 'GUEST'])
examples.add('loaded', 'show list of loaded plugins', 'loaded')

## statusline command

def handle_statusline(bot, event):
    """ show a status line. """
    event.reply("<b>controlchars:</b> %s - <b>perms:</b> %s" % (event.chan.data.cc, ", ".join(event.user.data.perms)))
	
cmnds.add('statusline', handle_statusline, ['USER', 'GUEST'])
examples.add('statusline', 'show status line', 'statusline')

## topper command

def handle_topper(bot, event):
    """ show a 'topper' startus line. """
    event.reply("<b>forwards:</b> %s - <b>watched:</b> %s  - <b>feeds:</b> %s" % (", ".join(event.chan.data.forwards) or "none", ", ".join(event.chan.data.watched) or "none", ", ".join([unicode(x) for x in event.chan.data.feeds]) or "none"))

cmnds.add('topper', handle_topper, ['USER', 'GUEST'])
examples.add('topper', 'show topper line', 'topper')

## running command

def handle_running(bot, event):
    """ show running tasks. """
    event.reply("<b>callbacks:</b> %s - <b>commands:</b> %s - <b>longrunning:</b> %s" % (defaultrunner.running(), cmndrunner.running(), longrunner.running()))

cmnds.add('running', handle_running, ['USER', 'GUEST'])
examples.add('running', "show running tasks", "running")



# me2/utils/url.py
#
# most code taken from maze

""" url related functions. """

## me2 imports

from generic import fromenc, toenc

## basic imports

import logging
import time
import sys
import re
import traceback
import Queue
import urllib
import urllib2
import urlparse
import socket
import random
import os
import sgmllib
import thread
import types
import httplib
import StringIO
import htmlentitydefs
import tempfile
import cgi

## defines

re_url_match  = re.compile(u'((?:http|https)://\S+)')

try: import chardet
except ImportError: chardet = None

class istr(str):
    pass

## useragent function

def useragent():
    """ provide useragent string """
    from me2.lib.version import version
    (name, version) = version.split()[0:2]
    return 'Mozilla/5.0 (compatible; %s %s; http://botme2.appspot.com)' % (name, version)

## CBURLopener class

class CBURLopener(urllib.FancyURLopener):
    """ our URLOpener """
    def __init__(self, version, *args):
        if version: self.version = version
        else: self.version = useragent()
        urllib.FancyURLopener.__init__(self, *args)

## geturl function

def geturl(url, version=None):
    """ fetch an url. """
    urllib._urlopener = CBURLopener(version)
    logging.warn('fetching %s' % url)
    result = urllib.urlopen(url)
    tmp = result.read()
    result.close()
    return tmp

## geturl2 function

def geturl2(url, decode=False):
    """ use urllib2 to fetch an url. """
    logging.warn('fetching %s' % url)
    request = urllib2.Request(url)
    request.add_header('User-Agent', useragent())
    opener = urllib2.build_opener()
    result = opener.open(request)
    tmp = result.read()
    info = result.info()
    result.close()
    if decode:
        encoding = get_encoding(tmp)
        logging.info('%s encoding: %s' % (url, encoding))
        res = istr(fromenc(tmp, encoding, url))
    else: res = istr(tmp)
    res.info = info
    return res

## geturl4 function

def geturl4(url, myheaders={}, postdata={}, keyfile="", certfile="", port=80):
    """ use httplib to fetch an url. """
    headers = {'Content-Type': 'text/html', 'Accept': 'text/plain; text/html', 'User-Agent': useragent()}
    headers.update(myheaders)
    urlparts = urlparse.urlparse(url)
    try:
       port = int(urlparts[1].split(':')[1])
       host = urlparts[1].split(':')[0]
    except: host = urlparts[1]
    if keyfile: connection = httplib.HTTPSConnection(host, port, keyfile, certfile)
    elif 'https' in urlparts[0]: connection = httplib.HTTPSConnection(host, port)
    else: connection = httplib.HTTPConnection(host, port)
    if type(postdata) == types.DictType: postdata = urllib.urlencode(postdata)
    logging.warn('fetching %s' % url)
    connection.request('GET', urlparts[2])
    return connection.getresponse()


## posturl function

def posturl(url, myheaders, postdata, keyfile=None, certfile="",port=80):
    """ very basic HTTP POST url retriever. """
    headers = {'Content-Type': 'application/x-www-form-urlencoded', 'Accept': 'text/plain; text/html', 'User-Agent': useragent()}
    headers.update(myheaders)
    urlparts = urlparse.urlparse(url)
    if keyfile: connection = httplib.HTTPSConnection(urlparts[1], port, keyfile, certfile)
    else: connection = httplib.HTTPConnection(urlparts[1])
    if type(postdata) == types.DictType: postdata = urllib.urlencode(postdata)
    logging.warn('fetching %s' % url)
    connection.request('POST', urlparts[2], postdata, headers)
    return connection.getresponse()

## delete url function

def deleteurl(url, myheaders={}, postdata={}, keyfile="", certfile="", port=80):
    """ very basic HTTP DELETE. """
    headers = {'Content-Type': 'application/x-www-form-urlencoded', 'Accept': 'text/plain; text/html', 'User-Agent': useragent()}
    headers.update(myheaders)
    urlparts = urlparse.urlparse(url)
    if keyfile and certfile: connection = httplib.HTTPSConnection(urlparts[1], port, keyfile, certfile)
    else: connection = httplib.HTTPConnection(urlparts[1])
    if type(postdata) == types.DictType: postdata = urllib.urlencode(postdata)
    logging.warn('fetching %s' % url)
    connection.request('DELETE', urlparts[2], postdata, headers)
    return connection.getresponse()

## put url function

def puturl(url, myheaders={}, postdata={}, keyfile="", certfile="", port=80):
    """ very basic HTTP PUT. """
    headers = {'Content-Type': 'application/x-www-form-urlencoded', 'Accept': 'text/plain; text/html', 'User-Agent': useragent()}
    headers.update(myheaders)
    urlparts = urlparse.urlparse(url)
    if keyfile: connection = httplib.HTTPSConnection(urlparts[1], port, keyfile, certfile)
    else: connection = httplib.HTTPConnection(urlparts[1])
    if type(postdata) == types.DictType: postdata = urllib.urlencode(postdata)
    logging.warn('fetching %s' % url)
    connection.request('PUT', urlparts[2], postdata, headers)
    return connection.getresponse()

## getpostdata function

def getpostdata(event):
    """ retrive post data from url data. """
    try:
        ctype, pdict = cgi.parse_header(event.headers.getheader('content-type'))
    except AttributeError: ctype, pdict = cgi.parse_header(event.headers.get('content-type'))
    body = cgi.FieldStorage(fp=event.rfile, headers=event.headers, environ = {'REQUEST_METHOD':'POST'}, keep_blank_values = 1)
    result = {}
    for name in dict(body): result[name] = body.getfirst(name)
    return result

## decode_html_entities function

def decode_html_entities(s):
    """ smart decoding of html entities to utf-8 """
    re_ent_match = re.compile(u'&([^;]+);')
    re_entn_match = re.compile(u'&#([^;]+);')
    s = s.decode('utf-8', 'replace')

    def to_entn(match):
        """ convert to entities """
        if htmlentitydefs.entitydefs.has_key(match.group(1)):
            return htmlentitydefs.entitydefs[match.group(1)].decode('utf-8', 'replace')
        return match.group(0)

    def to_utf8(match):
        """ convert to utf-8 """
        return unichr(long(match.group(1)))

    s = re_ent_match.sub(to_entn, s)
    s = re_entn_match.sub(to_utf8, s)
    return s

## get_encoding function

def get_encoding(data):
    """ get encoding from web data """
    if hasattr(data, 'info') and data.info.has_key('content-type') and 'charset' in data.info['content-type'].lower():
        charset = data.info['content-type'].lower().split('charset', 1)[1].strip()
        if charset[0] == '=':
            charset = charset[1:].strip()
            if ';' in charset: return charset.split(';')[0].strip()
            return charset
    if '<meta' in data.lower():
        metas = re.findall(u'<meta[^>]+>', data, re.I | re.M)
        if metas:
            for meta in metas:
                test_http_equiv = re.search('http-equiv\s*=\s*[\'"]([^\'"]+)[\'"]', meta, re.I)
                if test_http_equiv and test_http_equiv.group(1).lower() == 'content-type':
                    test_content = re.search('content\s*=\s*[\'"]([^\'"]+)[\'"]', meta, re.I)
                    if test_content:
                        test_charset = re.search('charset\s*=\s*([^\s\'"]+)', meta, re.I)
                        if test_charset: return test_charset.group(1)
    if chardet:
        test = chardet.detect(data)
        if test.has_key('encoding'): return test['encoding']
    return sys.getdefaultencoding()

## Stripper class

class Stripper(sgmllib.SGMLParser):

    """ html stripper. """

    def __init__(self):
        sgmllib.SGMLParser.__init__(self)

    def strip(self, some_html):
        """ strip html. """
        self.theString = u""
        self.feed(some_html)
        self.close()
        return self.theString

    def handle_data(self, data):
        """ data handler. """
        self.theString += data

## striphtml function

def striphtml(txt):
    """ strip html from txt. """
    stripper = Stripper()
    txt = stripper.strip(txt)
    return txt



# me2/plugs/misc.py
#
#

""" misc commands. """

## gozerbot imports

from me2.utils.exception import handle_exception
from me2.lib.commands import cmnds
from me2.lib.examples import examples
from me2.lib.persiststate import UserState
from me2.lib.socklib.utils.generic import waitforuser

## basic imports

import time
import os
import threading
import thread
import copy

## defines

cpy = copy.deepcopy

## test command

def handle_test(bot, ievent):
    """ give test response. """
    ievent.reply("%s - %s - it works!" % (ievent.auth or ievent.userhost, ievent.user.data.name))
    
cmnds.add('test', handle_test, ['USER', 'GUEST'])
examples.add('test', 'give test response',' test')

## source command

def handle_source(bot, ievent):
    """ show where to fetch the bot source. """ 
    ievent.reply('see http://botme2.googlecode.com')

cmnds.add('source', handle_source, ['USER', 'GUEST'])
examples.add('source', 'show source url', 'source')

## time command

def handle_time(bot, ievent):
    """ show current time """
    ievent.reply("time is %s" % time.ctime(time.time()))

cmnds.add('time', handle_time, ['USER', 'GUEST'])
examples.add('time', 'show current time', 'time')



# me2/tick.py
#
#

""" provide system wide clock tick. """

## me2 imports

from me2.lib.threadloop import TimedLoop
from me2.lib.eventbase import EventBase
from me2.lib.callbacks import callbacks

## TickLoop class

class TickLoop(TimedLoop):

    event = EventBase()
    event.type = event.cbtype = 'TICK'

    def start(self, bot=None):
        """ start the loop. """
        self.bot = bot
        TimedLoop.start(self)

    def handle(self):
        """ send TICK events to callback. """
        callbacks.check(self.bot, self.event)

## global tick loop

tickloop = TickLoop('tickloop', 60)



# handler_hubbub.py
#
#

## me2 imports

from me2.contrib import feedparser
from me2.lib.version import getversion
from me2.lib.plugins import plugs

## google imports

from google.appengine.api import urlfetch
from google.appengine.api import xmpp
from google.appengine.ext import db
from google.appengine.ext import webapp
from google.appengine.ext.webapp.util import run_wsgi_app
from google.appengine.ext.webapp import xmpp_handlers

## basic imports

import base64
import logging
import urllib
import urlparse
import uuid

logging.warn(getversion('GAE HUBBUB'))

if not plugs.has_key("me2.plugs.common.hubbub"):
    p = plugs.load("me2.plugs.common.hubbub")
else:
    p = plugs["me2.plugs.common.hubbub"]

class CallbackHandler(webapp.RequestHandler):

  def get(self):
    logging.warn('hubbub - incoming GET')
    if self.request.GET['hub.mode'] == 'unsubscribe':
      self.response.headers['Content-Type'] = 'text/plain'
      self.response.out.write(self.request.GET['hub.challenge'])
      return
      
    if self.request.GET['hub.mode'] != 'subscribe':
      self.error(400)
      return

    self.response.headers['Content-Type'] = 'text/plain'
    self.response.out.write(self.request.GET['hub.challenge'])

  def post(self):

    """Handles new content notifications."""

    logging.warn("hubbub - incoming POST")

    try:
        p.watcher.incoming(self.request.body)
    except IndexError:
        logging.error("hubbub plugin did not load properly")
    except Exception, ex:
        handle_exception()
        self.send_error(500)

application = webapp.WSGIApplication([('/(?:hubbub)', CallbackHandler)], debug=False)


def main():
  run_wsgi_app(application)


if __name__ == '__main__':
  main()



# me2/console/bot.py
#
#

""" console bot. """

## me2 imports

from me2.lib.datadir import getdatadir
from me2.utils.generic import waitforqueue
from me2.lib.errors import NoSuchCommand, NoInput
from me2.lib.botbase import BotBase
from me2.lib.exit import globalshutdown
from me2.utils.generic import strippedtxt
from me2.utils.exception import handle_exception
from event import ConsoleEvent

## basic imports

import time
import Queue
import logging
import sys
import code
import os
import readline
import atexit
import getpass
import re

## defines

histfilepath = os.path.expanduser(getdatadir() + os.sep + "run" + os.sep + "console-history")

## HistoryConsole class

class HistoryConsole(code.InteractiveConsole):
    def __init__(self, locals=None, filename="<console>", histfile=histfilepath):
        self.fname = histfile
        code.InteractiveConsole.__init__(self, locals, filename)
        self.init_history(histfile)

    def init_history(self, histfile):
        readline.parse_and_bind("tab: complete")
        if hasattr(readline, "read_history_file"):
            try: readline.read_history_file(histfile)
            except IOError: pass

    def save_history(self, histfile=None):
        readline.write_history_file(histfile or self.fname)

## the console

console = HistoryConsole()

## ConsoleBot

class ConsoleBot(BotBase):

    ERASE_LINE = '\033[2K'
    BOLD='\033[1m'
    RED = '\033[91m'
    YELLOW = '\033[93m'
    GREEN = '\033[92m'
    ENDC = '\033[0m'


    def __init__(self, cfg=None, users=None, plugs=None, botname=None, *args, **kwargs):
        BotBase.__init__(self, cfg, users, plugs, botname, *args, **kwargs)
        self.type = "console"
        self.nick = botname or "console"

    def start(self):
        """ start the console bot. """
        time.sleep(0.1)
        while not self.stopped: 
            try: 
                input = console.raw_input("> ")
                event = ConsoleEvent()
                event.parse(self, input, console)
                if False and input.startswith('#'):
                    try:
                        env = {"bot": self, "event": event}
                        env.update(locals())
                        env.update(globals())
                        console.locals.update(env)
                        console.runsource(input[1:])
                        continue
                    except Exception, ex:
                        handle_exception()
                        continue
                try:
                    #event.direct = True
                    result = self.doevent(event)
                    if not result: continue
                    logging.debug("console - waiting for %s to finish" % event.usercmnd)
                    res = waitforqueue(event.outqueue)
                    time.sleep(0.2)
                    logging.debug("console - %s" % res)
                except NoSuchCommand: print "no such command: %s" % event.usercmnd
            except NoInput: continue
            except (KeyboardInterrupt, EOFError): break
            except Exception, ex: handle_exception()
        console.save_history()


    def outnocb(self, printto, txt, *args, **kwargs):
        txt = self.normalize(txt)
        self._raw(txt)         

    def _raw(self, txt):
        """ do raw output to the console. """
        logging.info("%s - out - %s" % (self.name, txt))             
        #sys.stdout.write("\n")
        sys.stdout.write(txt)
        sys.stdout.write('\n')

    def action(self, channel, txt):
        txt = self.normalize(txt)
        self._raw(txt)

    def notice(self, channel, txt):
        txt = self.normalize(txt)
        self._raw(txt)

    def exit(self):
        """ called on exit. """
        console.save_history()

    def normalize(self, what):
        what = strippedtxt(what)
        what = what.replace("<b>", self.GREEN)
        what = what.replace("</b>", self.ENDC)
        what = what.replace("&lt;b&gt;", self.GREEN)
        what = what.replace("&lt;/b&gt;", self.ENDC)
        return what



#!/usr/bin/python2.4
#
# Copyright (C) 2009 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for the util module."""


__author__ = 'davidbyttow@google.com (David Byttow)'


import unittest

import ops
import util


class TestUtils(unittest.TestCase):
  """Tests utility functions."""

  def testIsIterable(self):
    self.assertTrue(util.is_iterable([]))
    self.assertTrue(util.is_iterable({}))
    self.assertTrue(util.is_iterable(set()))
    self.assertTrue(util.is_iterable(()))
    self.assertFalse(util.is_iterable(42))
    self.assertFalse(util.is_iterable('list?'))
    self.assertFalse(util.is_iterable(object))

  def testIsDict(self):
    self.assertFalse(util.is_dict([]))
    self.assertTrue(util.is_dict({}))
    self.assertFalse(util.is_dict(set()))
    self.assertFalse(util.is_dict(()))
    self.assertFalse(util.is_dict(42))
    self.assertFalse(util.is_dict('dict?'))
    self.assertFalse(util.is_dict(object))

  def testIsUserDefinedNewStyleClass(self):
    class OldClass:
      pass

    class NewClass(object):
      pass

    self.assertFalse(util.is_user_defined_new_style_class(OldClass()))
    self.assertTrue(util.is_user_defined_new_style_class(NewClass()))
    self.assertFalse(util.is_user_defined_new_style_class({}))
    self.assertFalse(util.is_user_defined_new_style_class(()))
    self.assertFalse(util.is_user_defined_new_style_class(42))
    self.assertFalse(util.is_user_defined_new_style_class('instance?'))

  def testLowerCamelCase(self):
    self.assertEquals('foo', util.lower_camel_case('foo'))
    self.assertEquals('fooBar', util.lower_camel_case('foo_bar'))
    self.assertEquals('fooBar', util.lower_camel_case('fooBar'))
    self.assertEquals('blipId', util.lower_camel_case('blip_id'))
    self.assertEquals('fooBar', util.lower_camel_case('foo__bar'))
    self.assertEquals('fooBarBaz', util.lower_camel_case('foo_bar_baz'))
    self.assertEquals('f', util.lower_camel_case('f'))
    self.assertEquals('f', util.lower_camel_case('f_'))
    self.assertEquals('', util.lower_camel_case(''))
    self.assertEquals('', util.lower_camel_case('_'))
    self.assertEquals('aBCDEF', util.lower_camel_case('_a_b_c_d_e_f_'))

  def assertListsEqual(self, a, b):
    self.assertEquals(len(a), len(b))
    for i in range(len(a)):
      self.assertEquals(a[i], b[i])

  def assertDictsEqual(self, a, b):
    self.assertEquals(len(a.keys()), len(b.keys()))
    for k, v in a.iteritems():
      self.assertEquals(v, b[k])

  def testSerializeList(self):
    data = [1, 2, 3]
    output = util.serialize(data)
    self.assertListsEqual(data, output)

  def testSerializeDict(self):
    data = {'key': 'value', 'under_score': 'value2'}
    expected = {'key': 'value', 'underScore': 'value2'}
    output = util.serialize(data)
    self.assertDictsEqual(expected, output)

  def testNonNoneDict(self):
    a = {'a': 1, 'b': 1}
    self.assertDictsEqual(a, util.non_none_dict(a))
    b = a.copy()
    b['c'] = None
    self.assertDictsEqual(a, util.non_none_dict(b))

  def testForceUnicode(self):
    self.assertEquals(u"aaa", util.force_unicode("aaa"))
    self.assertEquals(u"12", util.force_unicode(12))
    self.assertEquals(u"\u0430\u0431\u0432",
                      util.force_unicode("\xd0\xb0\xd0\xb1\xd0\xb2"))
    self.assertEquals(u'\u30e6\u30cb\u30b3\u30fc\u30c9',
                      util.force_unicode(u'\u30e6\u30cb\u30b3\u30fc\u30c9'))

  def testSerializeAttributes(self):

    class Data(object):
      def __init__(self):
        self.public = 1
        self._protected = 2
        self.__private = 3

      def Func(self):
        pass

    data = Data()
    output = util.serialize(data)
    # Functions and non-public fields should not be serialized.
    self.assertEquals(1, len(output.keys()))
    self.assertEquals(data.public, output['public'])

  def testStringEnum(self):
    util.StringEnum()
    single = util.StringEnum('foo')
    self.assertEquals('foo', single.foo)
    multi = util.StringEnum('foo', 'bar')
    self.assertEquals('foo', multi.foo)
    self.assertEquals('bar', multi.bar)

  def testParseMarkup(self):
    self.assertEquals('foo', util.parse_markup('foo'))
    self.assertEquals('foo bar', util.parse_markup('foo <b>bar</b>'))
    self.assertEquals('foo\nbar', util.parse_markup('foo<br>bar'))
    self.assertEquals('foo\nbar', util.parse_markup('foo<p indent="3">bar'))

if __name__ == '__main__':
  unittest.main()



#!/usr/bin/python2.4
#
# Copyright (C) 2009 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for the blip module."""


import unittest

import blip
import element
import ops
import simplejson

TEST_BLIP_DATA = {
    'childBlipIds': [],
    'content': '\nhello world!\nanother line',
    'contributors': ['robot@test.com', 'user@test.com'],
    'creator': 'user@test.com',
    'lastModifiedTime': 1000,
    'parentBlipId': None,
    'annotations': [{'range': {'start': 2, 'end': 3},
                     'name': 'key', 'value': 'val'}],
    'waveId': 'test.com!w+g3h3im',
    'waveletId': 'test.com!root+conv',
    'elements':{'14':{'type':'GADGET','properties':{'url':'http://a/b.xml'}}},
}

CHILD_BLIP_ID = 'b+42'
ROOT_BLIP_ID = 'b+43'


class TestBlip(unittest.TestCase):
  """Tests the primary data structures for the wave model."""

  def assertBlipStartswith(self, expected, totest):
    actual = totest.text[:len(expected)]
    self.assertEquals(expected, actual)

  def new_blip(self, **args):
    """Create a blip for testing."""
    data = TEST_BLIP_DATA.copy()
    data.update(args)
    res = blip.Blip(data, self.all_blips, self.operation_queue)
    self.all_blips[res.blip_id] = res
    return res

  def setUp(self):
    self.all_blips = {}
    self.operation_queue = ops.OperationQueue()

  def testBlipProperties(self):
    root = self.new_blip(blipId=ROOT_BLIP_ID,
                         childBlipIds=[CHILD_BLIP_ID])
    child = self.new_blip(blipId=CHILD_BLIP_ID,
                          parentBlipId=ROOT_BLIP_ID)
    self.assertEquals(ROOT_BLIP_ID, root.blip_id)
    self.assertEquals(set([CHILD_BLIP_ID]), root.child_blip_ids)
    self.assertEquals(set(TEST_BLIP_DATA['contributors']), root.contributors)
    self.assertEquals(TEST_BLIP_DATA['creator'], root.creator)
    self.assertEquals(TEST_BLIP_DATA['content'], root.text)
    self.assertEquals(TEST_BLIP_DATA['lastModifiedTime'],
                      root.last_modified_time)
    self.assertEquals(TEST_BLIP_DATA['parentBlipId'], root.parent_blip_id)
    self.assertEquals(TEST_BLIP_DATA['waveId'], root.wave_id)
    self.assertEquals(TEST_BLIP_DATA['waveletId'], root.wavelet_id)
    self.assertEquals(TEST_BLIP_DATA['content'][3], root[3])
    self.assertEquals(element.Gadget.class_type, root[14].type)
    self.assertEquals('http://a/b.xml', root[14].url)
    self.assertEquals('a', root.text[14])
    self.assertEquals(len(TEST_BLIP_DATA['content']), len(root))
    self.assertTrue(root.is_root())
    self.assertFalse(child.is_root())
    self.assertEquals(root, child.parent_blip)

  def testBlipSerialize(self):
    root = self.new_blip(blipId=ROOT_BLIP_ID,
                         childBlipIds=[CHILD_BLIP_ID])
    serialized = root.serialize()
    unserialized = blip.Blip(serialized, self.all_blips, self.operation_queue)
    self.assertEquals(root.blip_id, unserialized.blip_id)
    self.assertEquals(root.child_blip_ids, unserialized.child_blip_ids)
    self.assertEquals(root.contributors, unserialized.contributors)
    self.assertEquals(root.creator, unserialized.creator)
    self.assertEquals(root.text, unserialized.text)
    self.assertEquals(root.last_modified_time, unserialized.last_modified_time)
    self.assertEquals(root.parent_blip_id, unserialized.parent_blip_id)
    self.assertEquals(root.wave_id, unserialized.wave_id)
    self.assertEquals(root.wavelet_id, unserialized.wavelet_id)
    self.assertTrue(unserialized.is_root())

  def testDocumentOperations(self):
    blip = self.new_blip(blipId=ROOT_BLIP_ID)
    newlines = [x for x in blip.find('\n')]
    self.assertEquals(2, len(newlines))
    blip.first('world').replace('jupiter')
    bits = blip.text.split('\n')
    self.assertEquals(3, len(bits))
    self.assertEquals('hello jupiter!', bits[1])
    blip.range(2, 5).delete()
    self.assertBlipStartswith('\nho jupiter', blip)

    blip.first('ho').insert_after('la')
    self.assertBlipStartswith('\nhola jupiter', blip)
    blip.at(3).insert(' ')
    self.assertBlipStartswith('\nho la jupiter', blip)

  def testElementHandling(self):
    blip = self.new_blip(blipId=ROOT_BLIP_ID)
    url = 'http://www.test.com/image.png'

    org_len = len(blip)
    blip.append(element.Image(url=url))

    elems = [elem for elem in blip.find(element.Image, url=url)]
    self.assertEquals(1, len(elems))
    elem = elems[0]
    self.assertTrue(isinstance(elem, element.Image))
    blip.at(1).insert('twelve chars')
    self.assertTrue(blip.text.startswith('\ntwelve charshello'))

    elem = blip[org_len + 12].value()
    self.assertTrue(isinstance(elem, element.Image))

    blip.first('twelve ').delete()
    self.assertTrue(blip.text.startswith('\nchars'))

    elem = blip[org_len + 12 - len('twelve ')].value()
    self.assertTrue(isinstance(elem, element.Image))

    blip.first('chars').replace(element.Image(url=url))
    elems = [elem for elem in blip.find(element.Image, url=url)]
    self.assertEquals(2, len(elems))
    self.assertTrue(blip.text.startswith('\n hello'))
    elem = blip[1].value()
    self.assertTrue(isinstance(elem, element.Image))

  def testAnnotationHandling(self):
    key = 'style/fontWeight'

    def get_bold():
      for an in blip.annotations[key]:
        if an.value == 'bold':
          return an
      return None

    json = ('[{"range":{"start":3,"end":6},"name":"%s","value":"bold"}]'
            % key)
    blip = self.new_blip(blipId=ROOT_BLIP_ID,
                         annotations=simplejson.loads(json))
    self.assertEquals(1, len(blip.annotations))
    self.assertNotEqual(None, get_bold().value)
    self.assertTrue(key in blip.annotations)

    # extend the bold annotation by adding:
    blip.range(5, 8).annotate(key, 'bold')
    self.assertEquals(1, len(blip.annotations))
    self.assertEquals(8, get_bold().end)

    # clip by adding a same keyed:
    blip[4:12].annotate(key, 'italic')
    self.assertEquals(2, len(blip.annotations[key]))
    self.assertEquals(4, get_bold().end)

    # now split the italic one:
    blip.range(6, 7).clear_annotation(key)
    self.assertEquals(3, len(blip.annotations[key]))

    # test names and iteration
    self.assertEquals(1, len(blip.annotations.names()))
    self.assertEquals(3, len([x for x in blip.annotations]))
    blip[3: 5].annotate('foo', 'bar')
    self.assertEquals(2, len(blip.annotations.names()))
    self.assertEquals(4, len([x for x in blip.annotations]))
    blip[3: 5].clear_annotation('foo')

    # clear the whole thing
    blip.all().clear_annotation(key)
    # getting to the key should now throw an exception
    self.assertRaises(KeyError, blip.annotations.__getitem__, key)

  def testBlipOperations(self):
    blip = self.new_blip(blipId=ROOT_BLIP_ID)
    self.assertEquals(1, len(self.all_blips))

    otherblip = blip.reply()
    otherblip.append('hello world')
    self.assertEquals('hello world', otherblip.text)
    self.assertEquals(blip.blip_id, otherblip.parent_blip_id)
    self.assertEquals(2, len(self.all_blips))

    inline = blip.insert_inline_blip(3)
    self.assertEquals(blip.blip_id, inline.parent_blip_id)
    self.assertEquals(3, len(self.all_blips))

  def testInsertInlineBlipCantInsertAtTheBeginning(self):
    blip = self.new_blip(blipId=ROOT_BLIP_ID)
    self.assertEquals(1, len(self.all_blips))
    self.assertRaises(IndexError, blip.insert_inline_blip, 0)
    self.assertEquals(1, len(self.all_blips))

  def testDocumentModify(self):
    blip = self.new_blip(blipId=ROOT_BLIP_ID)
    blip.all().replace('a text with text and then some text')
    blip[7].insert('text ')
    blip.all('text').replace('thing')
    self.assertEquals('a thing thing with thing and then some thing',
                      blip.text)

  def testIteration(self):
    blip = self.new_blip(blipId=ROOT_BLIP_ID)
    blip.all().replace('aaa 012 aaa 345 aaa 322')
    count = 0
    prev = -1
    for start, end in blip.all('aaa'):
      count += 1
      self.assertTrue(prev < start)
      prev = start
    self.assertEquals(3, count)


  def testBlipRefValue(self):
    blip = self.new_blip(blipId=ROOT_BLIP_ID)
    content = blip.text
    content = content[:4] + content[5:]
    del blip[4]
    self.assertEquals(content, blip.text)

    content = content[:2] + content[3:]
    del blip[2:3]
    self.assertEquals(content, blip.text)

    blip[2:3] = 'bike'
    content = content[:2] + 'bike' + content[3:]
    self.assertEquals(content, blip.text)

    url = 'http://www.test.com/image.png'
    blip.append(element.Image(url=url))
    self.assertEqual(url, blip.first(element.Image).url)

    url2 = 'http://www.test.com/another.png'
    blip[-1].update_element({'url': url2})
    self.assertEqual(url2, blip.first(element.Image).url)

    self.assertTrue(blip[3:5] == blip.text[3:5])

    blip.append('geheim')
    self.assertTrue(blip.first('geheim'))
    self.assertFalse(blip.first(element.Button))
    blip.append(element.Button(name='test1', value='Click'))
    button = blip.first(element.Button)
    button.update_element({'name': 'test2'})
    self.assertEqual('test2', button.name)

  def testReplace(self):
    blip = self.new_blip(blipId=ROOT_BLIP_ID)
    blip.all().replace('\nxxxx')
    blip.all('yyy').replace('zzz')
    self.assertEqual('\nxxxx', blip.text)

  def testDeleteRangeThatSpansAcrossAnnotationEndPoint(self):
    json = ('[{"range":{"start":1,"end":3},"name":"style","value":"bold"}]')
    blip = self.new_blip(blipId=ROOT_BLIP_ID,
                         annotations=simplejson.loads(json),
                         content='\nFoo bar.')
    blip.range(2, 4).delete()
    self.assertEqual('\nF bar.', blip.text)
    self.assertEqual(1, blip.annotations['style'][0].start)
    self.assertEqual(2, blip.annotations['style'][0].end)

  def testInsertBeforeAnnotationStartPoint(self):
    json = ('[{"range":{"start":4,"end":9},"name":"style","value":"bold"}]')
    blip = self.new_blip(blipId=ROOT_BLIP_ID,
                         annotations=simplejson.loads(json),
                         content='\nFoo bar.')
    blip.at(4).insert('d and')
    self.assertEqual('\nFood and bar.', blip.text)
    self.assertEqual(9, blip.annotations['style'][0].start)
    self.assertEqual(14, blip.annotations['style'][0].end)

  def testDeleteRangeInsideAnnotation(self):
    json = ('[{"range":{"start":1,"end":5},"name":"style","value":"bold"}]')
    blip = self.new_blip(blipId=ROOT_BLIP_ID,
                         annotations=simplejson.loads(json),
                         content='\nFoo bar.')
    blip.range(2, 4).delete()

    self.assertEqual('\nF bar.', blip.text)
    self.assertEqual(1, blip.annotations['style'][0].start)
    self.assertEqual(3, blip.annotations['style'][0].end)

  def testReplaceInsideAnnotation(self):
    json = ('[{"range":{"start":1,"end":5},"name":"style","value":"bold"}]')
    blip = self.new_blip(blipId=ROOT_BLIP_ID,
                         annotations=simplejson.loads(json),
                         content='\nFoo bar.')
    blip.range(2, 4).replace('ooo')
    self.assertEqual('\nFooo bar.', blip.text)
    self.assertEqual(1, blip.annotations['style'][0].start)
    self.assertEqual(6, blip.annotations['style'][0].end)

    blip.range(2, 5).replace('o')
    self.assertEqual('\nFo bar.', blip.text)
    self.assertEqual(1, blip.annotations['style'][0].start)
    self.assertEqual(4, blip.annotations['style'][0].end)

  def testReplaceSpanAnnotation(self):
    json = ('[{"range":{"start":1,"end":4},"name":"style","value":"bold"}]')
    blip = self.new_blip(blipId=ROOT_BLIP_ID,
                         annotations=simplejson.loads(json),
                         content='\nFoo bar.')
    blip.range(2, 9).replace('')
    self.assertEqual('\nF', blip.text)
    self.assertEqual(1, blip.annotations['style'][0].start)
    self.assertEqual(2, blip.annotations['style'][0].end)

  def testSearchWithNoMatchShouldNotGenerateOperation(self):
    blip = self.new_blip(blipId=ROOT_BLIP_ID)
    self.assertEqual(-1, blip.text.find(':('))
    self.assertEqual(0, len(self.operation_queue))
    blip.all(':(').replace(':)')
    self.assertEqual(0, len(self.operation_queue))

  def testBlipsRemoveWithId(self):
    blip_dict = {
        ROOT_BLIP_ID: self.new_blip(blipId=ROOT_BLIP_ID,
                                    childBlipIds=[CHILD_BLIP_ID]),
        CHILD_BLIP_ID: self.new_blip(blipId=CHILD_BLIP_ID,
                                     parentBlipId=ROOT_BLIP_ID)
    }
    blips = blip.Blips(blip_dict)
    blips._remove_with_id(CHILD_BLIP_ID)
    self.assertEqual(1, len(blips))
    self.assertEqual(0, len(blips[ROOT_BLIP_ID].child_blip_ids))

  def testAppendMarkup(self):
    blip = self.new_blip(blipId=ROOT_BLIP_ID, content='\nFoo bar.')
    markup = '<p><span>markup<span> content</p>'
    blip.append_markup(markup)
    self.assertEqual(1, len(self.operation_queue))
    self.assertEqual('\nFoo bar.\nmarkup content', blip.text)

  def testBundledAnnotations(self):
    blip = self.new_blip(blipId=ROOT_BLIP_ID, content='\nFoo bar.')
    blip.append('not bold')
    blip.append('bold', bundled_annotations=[('style/fontWeight', 'bold')])
    self.assertEqual(2, len(blip.annotations))
    self.assertEqual('bold', blip.annotations['style/fontWeight'][0].value)

  def testInlineBlipOffset(self):
    offset = 14
    self.new_blip(blipId=ROOT_BLIP_ID,
                  childBlipIds=[CHILD_BLIP_ID],
                  elements={str(offset):
                      {'type': element.Element.INLINE_BLIP_TYPE,
                       'properties': {'id': CHILD_BLIP_ID}}})
    child = self.new_blip(blipId=CHILD_BLIP_ID,
                          parentBlipId=ROOT_BLIP_ID)
    self.assertEqual(offset, child.inline_blip_offset)

if __name__ == '__main__':
  unittest.main()



# me2/users.py
#
#

""" bot's users in JSON file.  NOT USED AT THE MOMENT. """

## lib imports

from me2.utils.exception import handle_exception, exceptionmsg
from me2.utils.generic import stripped
from me2.utils.name import stripname
from persist import Persist
from me2.utils.lazydict import LazyDict
from datadir import getdatadir
from errors import NoSuchUser
from config import Config

## basic imports

import re
import types
import os
import time
import logging

## JsonUser class

class JsonUser(Persist):

    """ LazyDict representing a user. """

    def __init__(self, name, userhosts=[], perms=[], permits=[], status=[], email=[]):
        assert name
        name = stripname(name.lower())
        Persist.__init__(self, getdatadir() + os.sep + 'users' + os.sep + name)
        self.data.datadir = self.data.datadir or getdatadir()
        self.data.name = self.data.name or name
        self.data.userhosts = self.data.userhosts or userhosts
        self.data.perms = self.data.perms or perms
        self.data.permits = self.data.permits or permits
        self.data.status = self.data.status or status
        self.data.email = self.data.email or email

## Users class

class Users(Persist):

    """ class representing all users. """

    def __init__(self, ddir=None, filename=None):
        self.datadir = ddir or getdatadir()
        self.filename = filename or 'mainusers'
        Persist.__init__(self, self.datadir + os.sep + self.filename)
        if not self.data: self.data = LazyDict()
        self.data.names = self.data.names or {}

    def all(self):
        """ get all users. """
        result = []
        for name in self.data['names'].values(): result.append(JsonUser(name).lower())
        return result

    ## Misc. Functions

    def size(self):
        """ return nr of users. """
        return len(self.data['names'])

    def names(self):
        """ get names of all users. """
        return self.data.names

    def byname(self, name):
        """ return user by name. """ 
        try:
            name = name.lower()
            name = stripname(name)
            user = JsonUser(name)
            if user.data.userhosts: return user
        except KeyError: raise NoSuchUser(name)

    def merge(self, name, userhost):
        """ add userhosts to user with name """
        name = name.lower()
        user = self.byname(name)
        if user:
            if not userhost in user.data.userhosts:
                user.data.userhosts.append(userhost)
                user.save()
            self.data.names[userhost] = name
            self.save()
            logging.warn("users - %s merged with %s" % (userhost, name))
            return 1

    def usersearch(self, userhost):
        """ search for users with a userhost like the one specified """
        result = []
        for u, name in self.data.names.iteritems():
            if userhost in u: result.append((name.lower(), u))
        return result

    def getuser(self, userhost):
        """ get user based on userhost. """
        try:
            user = self.byname(self.data.names[userhost])
            if user: return user
        except KeyError:
            logging.debug("user - can't find %s in names cache" % userhost) 

    ## Check functions

    def exist(self, name):
        """ see if user with <name> exists """
        return self.byname(name)

    def allowed(self, userhost, perms, log=True, bot=None):
        """ check if user with userhosts is allowed to execute perm command """
        if not type(perms) == types.ListType: perms = [perms, ]
        if 'ANY' in perms: return 1
        if bot and bot.allowall: return 1
        res = None
        user = self.getuser(userhost)
        if not user:
            logging.warn('users - %s userhost denied' % userhost)
            return res
        else:
            uperms = set(user.data.perms)
            sperms = set(perms)
            intersection = sperms.intersection(uperms)
            res = list(intersection) or None
        if not res and log: logging.warn("users - %s perm %s denied (%s)" % (userhost, str(perms), str(uperms)))
        return res

    def permitted(self, userhost, who, what):
        """ check if (who,what) is in users permit list """
        user = self.getuser(userhost)
        res = None
        if user: 
            if '%s %s' % (who, what) in user.data.permits: res = 1
        return res

    def status(self, userhost, status):
        """ check if user with <userhost> has <status> set """
        user = self.getuser(userhost)
        res = None
        if user:
            if status.upper() in user.data.status: res = 1
        return res

    def gotuserhost(self, name, userhost):
        """ check if user has userhost """
        user = self.byname(name)
        return userhost in user.data.userhosts

    def gotperm(self, name, perm):
        """ check if user had permission """
        user = self.byname(name)
        if user: return perm.upper() in user.data.perms

    def gotpermit(self, name, permit):
        """ check if user permits something.  permit is a (who, what) tuple """
        user = self.byname(name)
        if user: return '%s %s' % permit in user.data.permits
        
    def gotstatus(self, name, status):
        """ check if user has status """
        user = self.byname(name)
        return status.upper() in user.data.status

    ## Get Functions

    def getname(self, userhost):
        """ get name of user belonging to <userhost> """
        try: return self.data.names[userhost]
        except KeyError:
            try: return self.data.names[userhost]
            except KeyError:
                user = self.getuser(userhost)
                if user: return user.data.name.lower()

    def gethosts(self, userhost):
        """ return the userhosts of the user associated with the specified userhost """
        user = self.getuser(userhost)
        if user: return user.data.userhosts
    
    def getemail(self, userhost):
        """ return the email of the specified userhost """
        user = self.getuser(userhost)
        if user:
            if user.data.email: return user.data.email[0]

    def getperms(self, userhost):
        """ return permission of user"""
        user = self.getuser(userhost)
        if user: return user.data.perms

    def getpermits(self, userhost):
        """ return permits of the specified userhost"""
        user = self.getuser(userhost)
        if user: return user.data.permits

    def getstatuses(self, userhost):
        """ return the list of statuses for the specified userhost. """
        user = self.getuser(userhost)
        if user: return user.data.status

    def getuserhosts(self, name):
        """ return the userhosts associated with the specified user. """
        user = self.byname(name)
        if user: return user.data.userhosts

    def getuseremail(self, name):
        """ get email of user. """
        user = self.byname(name)
        if user:
            if user.data.email: return user.data.email[0]

    def getuserperms(self, name):
        """ return permission of user. """
        user = self.byname(name)
        if user: return user.data.perms

    def getuserpermits(self, name):
        """ return permits of user. """
        user = self.byname(name)
        if user: return user.data.permits

    def getuserstatuses(self, name):
        """ return the list of statuses for the specified user. """
        user = self.byname(name)
        if user: return user.data.status

    def getpermusers(self, perm):
        """ return all users that have the specified perm. """
        result = []
        for name in self.data.names:
            user = JsonUser(name)
            if perm.upper() in user.data.perms: result.append(user.data.name)
        return result

    def getstatususers(self, status):
        """ return all users that have the specified status. """
        result = []
        for name in self.data.names:
            user = JsonUser(name.lower())
            if status in user.data.status: result.append(user.data.name.lower())
        return result

    ## Set Functions

    def setemail(self, name, email):
        """ set email of user. """
        user = self.byname(name)
        if user:
            try: user.data.email.remove(email)
            except: pass
            user.data.email.insert(0, email)
            user.save()
            return True
        return False

    ## Add functions

    def add(self, name, userhosts, perms):
        """ add an user. """
        name = name.lower()
        newuser = JsonUser(name, userhosts, perms)
        for userhost in userhosts: self.data.names[userhost] = name
        newuser.save()
        self.save()
        logging.warn('users - %s %s %s added to user database' % (name, userhosts, perms))
        return True

    def addguest(self, userhost):
        if not self.getname(userhost):
            if Config().guestasuser: self.add(userhost, [userhost, ], ["USER",])
            else: self.add(userhost, [userhost, ], ["GUEST",])

    def addemail(self, userhost, email):
        """ add an email address to the userhost. """
        user = self.getuser(userhost)
        if user:
            user.data.email.append(email)
            user.save()
            return 1

    def addperm(self, userhost, perm):
        """ add the specified perm to the userhost. """
        user = self.getuser(userhost)
        if user:
            user.data.perms.append(perm.upper())
            user.save()
            return 1

    def addpermit(self, userhost, permit):
        """ add the given (who, what) permit to the given userhost. """
        user = self.getuser(userhost)
        if user:
            user.data.permits.append(permit)
            user.save()
            return 1

    def addstatus(self, userhost, status):
        """ add status to given userhost. """
        user = self.getuser(userhost)
        if user:
            user.data.status.append(status.upper())
            user.save()
            return 1

    def adduserhost(self, name, userhost):
        """ add userhost. """
        name = name.lower()
        user = self.byname(name)
        if not user: user = self.users[name] = JsonUser(name=name)
        user.data.userhosts.append(userhost)
        user.save()
        return 1

    def adduseremail(self, name, email):
        """ add email to specified user. """
        user = self.byname(name)
        if user:
            user.data.email.append(email)
            user.save()
            return 1

    def adduserperm(self, name, perm):
        """ add permission. """
        user = self.byname(name)
        if user:
            perm = perm.upper()
            user.data.perms.append(perm)
            user.save()
            return 1

    def adduserpermit(self, name, who, permit):
        """ add (who, what) permit tuple to sepcified user. """
        user = self.byname(name)
        if user:
            p = '%s %s' % (who, permit)
            user.data.permits.append(p)
            user.save()
            return 1

    def adduserstatus(self, name, status):
        """ add status to given user. """
        user = self.byname(name)
        if user:
            user.data.status.append(status.upper())
            user.save()
            return 1

    def addpermall(self, perm): 
        """ add permission to all users. """
        for name in self.data.names:
            user = JsonUser(name.lower())
            user.data.perms.append(perm.upper())
            user.save()

    ## Delete functions

    def delemail(self, userhost, email):
        """ delete email from userhost. """
        user = self.getuser(userhost)
        if user:
            if email in user.emails:
                user.data.emails.remove(email)
                user.save()
                return 1

    def delperm(self, userhost, perm):
        """ delete perm from userhost. """
        user = self.getuser(userhost)
        if user:
            p = perm.upper()
            if p in user.perms:
                user.data.perms.remove(p)
                user.save()
                return 1

    def delpermit(self, userhost, permit):
        """ delete permit from userhost. """
        user = self.getuser(userhost)
        if user:
            p = '%s %s' % permit
            if p in user.permits:
                user.data.permits.remove(p)
                user.save()
                return 1

    def delstatus(self, userhost, status):
        """ delete status from userhost. """
        user = self.getuser(userhost)
        if user:
            st = status.upper()
            if st in user.data.status:
                user.data.status.remove(st)
                user.save()
                return 1

    def delete(self, name):
        """ delete user with name. """
        try:
            name = name.lower()
            user = JsonUser(name)
            user.data.deleted = True
            user.save()
            if user:
                 for userhost in user.data.userhosts:
                     try: del self.data.names[userhost]
                     except KeyError: pass
            self.save()
            return True
        except NoSuchUser:
            pass
        
    def deluserhost(self, name, userhost):
        """ delete the userhost entry. """
        user = self.byname(name)
        if user:
            if userhost in user.data.userhosts:
                user.data.userhosts.remove(userhost)
                user.save()
                return 1

    def deluseremail(self, name, email):
        """ delete email. """
        user = self.byname(name)
        if user:
            if email in user.data.email:
                user.data.email.remove(email)
                user.save()
                return 1

    def deluserperm(self, name, perm):
        """ delete permission. """
        user = self.byname(name)
        if user:
            p = perm.upper()
            if p in user.data.perms:
                user.data.perms.remove(p)
                user.save()
                return 1

    def deluserpermit(self, name, permit):
        """ delete permit. """
        user = self.byname(name)
        if user:
            p = '%s %s' % permit
            if p in user.data.permits:
                user.data.permits.remove(p)
                user.save()
                return 1

    def deluserstatus(self, name, status):
        """ delete the status from the given user. """
        user = self.byname(name)
        if user:
            st = status.upper()
            if st in user.data.status:
                user.data.status.remove(status)
                user.save()
                return 1

    def delallemail(self, name):
        """ delete all emails for the specified user. """
        user = self.byname(name)
        if user:
            user.data.email = []
            user.save()
            return 1

    def make_owner(self, userhosts):
        """ see if owner already has a user account if not add it. """ 
        if not userhosts:
            logging.info("no usershosts provided in make_owner")
            return
        owner = []
        if type(userhosts) != types.ListType: owner.append(userhosts)
        else: owner = userhosts
        for userhost in owner:
            username = self.getname(unicode(userhost))
            if not username or username != 'owner':
                if not self.merge('owner', unicode(userhost)): self.add('owner', [unicode(userhost), ], ['USER', 'OPER'])

## global users object

users = None

## users_boot function

def users_boot():
    """ initialize global users object. """
    global users
    users = Users()
    return users



# me2.plugs.common/forward.py
#
#

""" forward incoming trafic on a bot to another bot through xmpp. """

## me2 imports

from me2.lib.commands import cmnds
from me2.lib.callbacks import callbacks, remote_callbacks, last_callbacks, first_callbacks
from me2.lib.eventbase import EventBase
from me2.lib.persist import PlugPersist
from me2.utils.lazydict import LazyDict
from me2.lib.examples import examples
from me2.lib.fleet import getfleet
from me2.lib.container import Container
from me2.lib.errors import NoProperDigest
from me2.utils.exception import handle_exception
from me2.utils.locking import locked
from me2.utils.generic import strippedtxt

## me2.plugs.common imports

from me2.plugs.common.twitter import postmsg

## xmpp import

from me2.contrib.xmlstream import NodeBuilder, XMLescape, XMLunescape


## basic imports

import logging
import copy
import time
import types
import hmac 
import hashlib

## simpljejson imports

from me2.contrib.simplejson import loads, dumps

## defines

forward = PlugPersist("forward-core")
if not forward.data.allowin:
    forward.data.allowin = []
if not forward.data.channels:
    forward.data.channels = {}
if not forward.data.outs:
    forward.data.outs = {}
if not forward.data.whitelist:
    forward.data.whitelist = {}

cpy = copy.deepcopy

## forward precondition

def forwardoutpre(bot, event):
    """ preconditon to check if forward callbacks is to be fired. """
    chan = unicode(event.channel).lower()
    if not chan: return
    logging.debug("forward - pre - %s" % event.channel)
    if chan in forward.data.channels and not event.isremote() and not event.forwarded:
        if event.how != u"background": return True
    return False

## forward callbacks

def forwardoutcb(bot, event): 
    """ forward callback. """
    e = cpy(event)
    logging.debug("forward - cbtype is %s - %s" % (event.cbtype, event.how))
    e.forwarded = True
    e.source = bot.jid
    e.botname = bot.server or bot.name
    if not event.chan: event.bind(bot)
    if event.chan: e.allowwatch = event.chan.data.allowwatch
    fleet = getfleet()
    for jid in forward.data.channels[event.channel.lower()]:
        logging.info("forward - sending to %s" % jid)
        if jid == "twitter":
            try: postmsg(forward.data.outs[jid], e.txt)
            except Exception, ex: handle_exception()
            continue
        outbot = fleet.getfirstjabber(bot.isgae)
        if not outbot and bot.isgae: outbot = fleet.makebot('xmpp', 'forwardbot')
        if outbot:
            e.source = outbot.jid
            txt = outbot.normalize(e.tojson())
            #txt = e.tojson()
            container = Container(outbot.jid, txt)
            outbot.outnocb(jid, container.tojson()) 
        else: logging.info("forward - no xmpp bot found in fleet".upper())

first_callbacks.add('BLIP_SUBMITTED', forwardoutcb, forwardoutpre)
first_callbacks.add('MESSAGE', forwardoutcb, forwardoutpre)
#first_callbacks.add('PRESENCE', forwardoutcb, forwardoutpre)
first_callbacks.add('PRIVMSG', forwardoutcb, forwardoutpre)
first_callbacks.add('JOIN', forwardoutcb, forwardoutpre)
first_callbacks.add('PART', forwardoutcb, forwardoutpre)
first_callbacks.add('QUIT', forwardoutcb, forwardoutpre)
first_callbacks.add('NICK', forwardoutcb, forwardoutpre)
first_callbacks.add('CONSOLE', forwardoutcb, forwardoutpre)
first_callbacks.add('WEB', forwardoutcb, forwardoutpre)
first_callbacks.add('DISPATCH', forwardoutcb, forwardoutpre)
first_callbacks.add('OUTPUT', forwardoutcb, forwardoutpre)

## forward-add command

def handle_forwardadd(bot, event):
    """ add a new forward. """
    if not event.rest:
        event.missing('<JID>')
        return
    if "@" in event.rest:
        forward.data.outs[event.rest] = event.user.data.name
        forward.save()
        if not event.rest in event.chan.data.forwards: event.chan.data.forwards.append(event.rest)
    if event.rest:
        event.chan.save()
        event.done()

cmnds.add("forward-add", handle_forwardadd, 'OPER')
examples.add("forward-add" , "add an JID to forward to", "forward-add jsoncloud@appspot.com")

## forward-del command

def handle_forwarddel(bot, event):
    """ delete a forward. """
    if not event.rest:
        event.missing('<JID>')
        return
    del forward.data.outs[event.rest]
    forward.save()
    event.done()

cmnds.add("forward-del", handle_forwarddel, 'OPER')
examples.add("forward-del" , "delete an JID to forward to", "forward-del jsoncloud@appspot.com")

## forward-allow command

def handle_forwardallow(bot, event):
    """ allow a remote bot to forward to us. """
    if not event.rest:
        event.missing("<JID>")
        return
    if forward.data.whitelist.has_key(event.rest):
        forward.data.whitelist[event.rest] = bot.type
        forward.save()
    event.done()

cmnds.add("forward-allow", handle_forwardallow, 'OPER')
examples.add("forward-allow" , "allow an JID to forward to us", "forward-allow jsoncloud@appspot.com")

## forward-list command

def handle_forwardlist(bot, event):
    """ list forwards. """
    try: event.reply("forwards for %s: " % event.channel, forward.data.channels[event.channel])
    except KeyError: event.reply("no forwards for %s" % event.channel)

cmnds.add("forward-list", handle_forwardlist, 'OPER')
examples.add("forward-list" , "list all forwards of a channel", "forward-list")

## forward command

def handle_forward(bot, event):
    """ forward the channel tot another bot. """
    if not event.args:
        event.missing("<JID>")
        return
    forward.data.channels[event.channel.lower()] =  event.args
    for jid in event.args:
        forward.data.outs[jid] = event.user.data.name
        if not jid in event.chan.data.forwards: event.chan.data.forwards = event.args
    if event.args: event.chan.save()
    forward.save()
    event.done()

cmnds.add("forward", handle_forward, 'OPER')
examples.add("forward" , "forward a channel to provided JIDS", "forward jsoncloud@appspot.com")

## forward-stop command

def handle_forwardstop(bot, event):
    """ stop forwarding the channel to another bot. """
    if not event.args:
        event.missing("<JID>")
        return

    try:
        for jid in event.args:
            try:
                forward.data.channels[event.channel].remove(jid)
                del forward.data.outs[jid]
                if jid in event.chan.data.forwards: event.chan.data.forwards.remove(jid)
            except ValueError: pass
        forward.save()
        event.done()
    except KeyError, ex: event.reply("we are not forwarding %s" %  str(ex))

cmnds.add("forward-stop", handle_forwardstop, 'OPER')
examples.add("forward-stop" , "stop forwarding a channel to provided JIDS", "forward-stop jsoncloud@appspot.com")



# me2.plugs.common/karma.py
#
#

## me2 imports

from me2.lib.callbacks import callbacks
from me2.lib.commands import cmnds
from me2.lib.examples import examples
from me2.lib.persist import PlugPersist
from me2.utils.statdict import StatDict

## basic imports

import logging

## KarmaItem class

class KarmaItem(PlugPersist):

     def __init__(self, name, default={}):
         PlugPersist.__init__(self, name)
         self.data.name = name
         self.data.count = self.data.count or 0
         self.data.whoup = self.data.whoup or {}
         self.data.whodown = self.data.whodown or {}
         self.data.whyup = self.data.whyup or []
         self.data.whydown = self.data.whydown or []

## karma precondition

def prekarma(bot, event):
    if event.iscmnd(): return True
    return False

## karma callbacks

def karmacb(bot, event):
    try: target = event.txt[1:].lower()
    except IndexError: return
    item = item2 = rest = reason = None
    try: item, rest = target.split("++", 1)
    except ValueError:
        try: item2, rest = target.split("--", 1)
        except ValueError: return
    try: reason = rest.split('#', 1)[1]
    except IndexError: reason = None
    if item:
        i = KarmaItem(event.channel + "-" + item)
        i.data.count += 1
        if event.nick not in i.data.whoup: i.data.whoup[event.nick] = 0
        i.data.whoup[event.nick] += 1
        if reason and reason not in i.data.whyup: i.data.whyup.append(reason)
        i.save()
    else:
        i = KarmaItem(event.channel + "-" + item2)
        i.data.count -= 1
        if event.nick not in i.data.whodown: i.data.whodown[event.nick] = 0
        i.data.whodown[event.nick] -= 1
        if reason and reason not in i.data.whyup: i.data.whydown.append(reason)
        i.save()
    got = item or item2
    event.bind(bot)
    event.reply("karma of %s is now %s" % (got, i.data.count))

callbacks.add('PRIVMSG', karmacb, prekarma)
callbacks.add('MESSAGE', karmacb, prekarma)

## karma command

def handle_karma(bot, event):
    if not event.rest: event.missing("<what>") ; return
    k = event.rest.lower()
    item = KarmaItem(event.channel + "-" + k)
    if item.data.count: event.reply("karma of %s is %s" % (k, item.data.count))
    else: event.reply("%s doesn't have karma yet." % k)

cmnds.add('karma', handle_karma, ['USER', ])
examples.add('karma', 'show karma', 'karma me2')

def handle_karmawhyup(bot, event):
    k = event.rest.lower()
    item = KarmaItem(event.channel + "-" + k)
    if item.data.whyup: event.reply("reasons for karma up are: ", item.data.whyup)
    else: event.reply("no reasons for karmaup of %s known yet" % k)

cmnds.add("karma-whyup", handle_karmawhyup, ['USER', 'GUEST'])
examples.add("karma-whyup", "show why a karma item is upped", "karma-whyup me2")

def handle_karmawhoup(bot, event):
    k = event.rest.lower()
    item = KarmaItem(event.channel + "-" + k)
    sd = StatDict(item.data.whoup)
    res = []
    for i in sd.top():
        res.append("%s: %s" % i)
    if item.data.whoup: event.reply("uppers of %s are: " % k, res)
    else: event.reply("nobody upped %s yet" % k)

cmnds.add("karma-whoup", handle_karmawhoup, ['USER', 'GUEST'])
examples.add("karma-whoup", "show who upped an item", "karma-whoup me2")

def handle_karmawhydown(bot, event):
    k = event.rest.lower()
    item = KarmaItem(event.channel + "-" + k)
    if item.data.whydown: event.reply("reasons for karmadown are: ", item.data.whydown)
    else: event.reply("no reasons for karmadown of %s known yet" % k)

cmnds.add("karma-whydown", handle_karmawhydown, ['USER', 'GUEST'])
examples.add("karma-whydown", "show why a karma item is downed", "karma-whydown me2")

def handle_karmawhodown(bot, event):
    k = event.rest.lower()
    item = KarmaItem(event.channel + "-" + k)
    sd = StatDict(item.data.whodown)
    res = []
    for i in sd.down():
        res.append("%s: %s" % i)
    if item.data.whodown: event.reply("downers of %s are: " % k, res)
    else: event.reply("nobody downed %s yet" % k)

cmnds.add("karma-whodown", handle_karmawhodown, ['USER', 'GUEST'])
examples.add("karma-whodown", "show who downed an item", "karma-whodown me2")



# me2/plugs/dispatch.py
#
#

""" base class for all bots. """

## me2 imports

from me2.lib.callbacks import last_callbacks
from me2.lib.errors import NoSuchCommand

## basic logging

import logging
import copy

## defines

cpy = copy.deepcopy


## dispatch precondition

def predispatch(bot, event):
    if event.status == "done":
        logging.debug("dispatch - event is done .. ignoring")
        return
    if event.isremote():
        logging.done("dispatch - event is remote .. not dispatching")
        return
    return True

## dispatch callback

def dispatch(bot, event):
    """ dispatch an event. """
    if event.nodispatch:
        logging.debug("dispatch - nodispatch option is set - ignoring %s" % event.userhost)
        return
    bot.status = "dispatch"
    event.bind(bot)
    bot.curevent = event
    go = False
    execstr = event.iscmnd()
    try:
        if execstr:
            event.iscommand = True
            e = cpy(event)
            e.usercmnd = execstr.split()[0]
            e.txt = execstr
            e.showexception = True
            if not e.options: e.makeoptions()
            e.prepare()
            if e.usercmnd in event.chan.data.silentcommands: e.silent = True
            result = bot.plugs.dispatch(bot, e)
            #event.leave()
        else:
            logging.debug("dispatch - no go for %s (cc is %s)" % (event.auth or event.userhost, execstr))
            result =  []
    except NoSuchCommand:
        logging.info("no such command: %s" % event.usercmnd)
        result = []
    return result

## register callback

last_callbacks.add('PRIVMSG', dispatch, predispatch)
last_callbacks.add('MESSAGE', dispatch, predispatch)
last_callbacks.add('BLIP_SUBMITTED', dispatch, predispatch)
last_callbacks.add('WEB', dispatch, predispatch)
last_callbacks.add('CONSOLE', dispatch, predispatch)
last_callbacks.add('DCC', dispatch, predispatch)
last_callbacks.add('DISPATCH', dispatch, predispatch)



# me2.plugs.common/learn.py
#
#

""" learn information items .. facts .. factoids. """

## me2 imports

from me2.lib.callbacks import callbacks
from me2.lib.commands import cmnds
from me2.lib.examples import examples
from me2.utils.lazydict import LazyDict
from me2.lib.persist import PlugPersist

## basic imports

import logging

## commands

def handle_learn(bot, event):
    """" set an information item. """
    if not event.rest: event.missing("<item> is <description>") ; return
    try: (what, description) = event.rest.split(" is ", 1)
    except ValueError: event.missing("<item> is <description>") ; return
    what = what.lower()
    items = PlugPersist(event.channel)
    if not items.data: items.data = LazyDict()
    if not items.data.has_key(what): items.data[what] = []
    if description not in items.data[what]: items.data[what].append(description)
    items.save()
    event.reply("%s item added to %s database" % (what, event.channel))

cmnds.add('learn', handle_learn, ['USER', 'GUEST'])
examples.add('learn', 'learn the bot a description of an item.', "learn dunk is botpapa")

def handle_forget(bot, event):
    """" set an information item. """
    if not event.rest: event.missing("<item> and <match>") ; return
    try: (what, match) = event.rest.split(" and ", 2)
    except ValueError: event.missing("<item> and <match>") ; return
    what = what.lower()
    items = PlugPersist(event.channel)
    if not items.data: items.data = LazyDict()
    if items.data.has_key(what):
        for i in range(len(items.data[what])):
            if match in items.data[what][i]:
                del items.data[what][i]                
                items.save()
                break
    event.reply("item removed from %s database" % event.channel)

cmnds.add('forget', handle_forget, ['USER'])
examples.add('forget', 'forget a description of an item.', "forget dunk and botpapa")

def handle_whatis(bot, event):
    items = PlugPersist(event.channel)
    what = event.rest.lower()
    if what in items.data and items.data[what]: event.reply("%s is " % event.rest, items.data[what], dot=", ")
    else: event.reply("no information known about %s" % what)

cmnds.add('whatis', handle_whatis, ['USER', 'GUEST'])
examples.add("whatis", "whatis learned about a subject", "whatis me2")

def prelearn(bot, event):
    if event.iscmnd() and not event.forwarded: return True
    return False

def learncb(bot, event):
    event.bind(bot)
    items = PlugPersist(event.channel)
    target = event.usercmnd[1:].lower()
    if target in items.data: event.reply("%s is " % target, items.data[target], dot=", ")

callbacks.add("PRIVMSG", learncb, prelearn)
callbacks.add("MESSAGE", learncb, prelearn)
callbacks.add("DISPATCH", learncb, prelearn)



# me2/utils/log.py
#
#

""" log module. """

## basic imports

import logging
import logging.handlers
import os
import os.path
import getpass

## defines

LEVELS = {'debug': logging.DEBUG,
          'info': logging.INFO,
          'warning': logging.WARNING,
          'warn': logging.WARNING,
          'error': logging.ERROR,
          'critical': logging.CRITICAL}

try:
    import waveapi
except ImportError:
    if os.path.isdir('/var/log/me2') and getpass.getuser() == "me2": LOGDIR = "/var/log/me2"
    else: LOGDIR = os.path.expanduser("~") + os.sep + ".me2" + os.sep + "botlogs"

try:
    ddir = os.sep.join(LOGDIR.split(os.sep)[:-1])
    if not os.path.isdir(ddir): os.mkdir(ddir)
except: pass

try:
    if not os.path.isdir(LOGDIR): os.mkdir(LOGDIR)
except: pass

format = "%(asctime)s - %(message)s - <%(threadName)s+%(module)s-%(funcName)s:%(lineno)s>"

try:
    import waveapi
except ImportError:
    filehandler = logging.handlers.TimedRotatingFileHandler(LOGDIR + os.sep + "me2.log", 'midnight')
    formatter = logging.Formatter(format)
    filehandler.setFormatter(formatter)  

## setloglevel function

def setloglevel(level_name):
    """ set loglevel to level_name. """
    level = LEVELS.get(level_name, logging.NOTSET)
    root = logging.getLogger("")
    if root.handlers:
        for handler in root.handlers: root.removeHandler(handler)
    logging.basicConfig(level=level, format=format)
    try: import waveapi
    except ImportError: root.addHandler(filehandler)
    root.setLevel(level)
    logging.info("loglevel is %s (%s)" % (str(level), level_name))



# handler_docs.py
#
#

""" xmpp request handler. """

## me2 imports

from me2.lib.version import getversion

## google imports

import webapp2

## basic imports

import sys
import time
import types
import logging

## greet

logging.warn(getversion('REDIRECT'))

## classes

class DocsHandler(webapp2.RequestHandler):

    def get(self, url=None):
        try:
            if not url.endswith(".html"):
                if not url.endswith('/'):
                    url += u"/index.html"
                else:
                    url += u"index.html"
            splitted = url.split('/')
            splitted.insert(2, 'html')
            goto = '/'.join(splitted)
            logging.warn("docs - redirecting %s" % goto)
            self.redirect(goto)
        except Exception, ex:
            handle_exception()
            #self.send_error(500)

application = webapp2.WSGIApplication([webapp2.Route(r'<url:.*>', DocsHandler)], 
                                      debug=True)

def main():
    global application
    application.run()

if __name__ == "__main__":
    main()



# me2.plugs.common/echo.py
#
#

""" echo typed sentences. """

## me2 imports

from me2.lib.commands import cmnds
from me2.lib.examples import examples
from me2.lib.callbacks import callbacks, first_callbacks

## basic imports

import logging

## echo callback

def echopre(bot, event):
    logging.warn("%s - echo pre %s" % (bot.name, event.cbtype))
    if event.how != "background": return True
    return False

def echocb(bot, event):
    if not event.isdcc:
        event.forwarded = True
        if event.cbtype == "OUTPUT": bot.outnocb(event.channel, u"[!] %s" % event.txt, event=event, dotime=True)
        else: bot.outnocb(event.channel, u"[%s] %s" % (event.nick, event.txt), event=event, dotime=True)

#callbacks.add("DISPATCH", echocb, echopre, threaded=True)
#first_callbacks.add("OUTPUT", echocb, echopre, threaded=True)

## echo command

def handle_echo(bot, event):
    """ echo txt to user. """
    if event.how != "background" and not event.iscmnd() and not event.isremote:
        if not event.isdcc: bot.saynocb(event.channel, u"[%s] %s" % (event.nick, event.txt))
            
cmnds.add("echo", handle_echo, ['USER', 'GUEST'])
examples.add("echo", "echo input", "echo yoooo dudes")



# me2/rest/client.py
#
#

""" Rest Client class """

## me2 imports

from me2.utils.url import geturl4, posturl, deleteurl, useragent
from me2.utils.generic import toenc
from me2.utils.exception import handle_exception, exceptionmsg
from me2.utils.locking import lockdec
from me2.utils.lazydict import LazyDict
from me2.contrib.simplejson import loads

## basic imports

from urllib2 import HTTPError, URLError
from httplib import InvalidURL
from urlparse import urlparse
import socket
import asynchat
import urllib
import sys
import thread
import re
import asyncore
import time
import logging

## defines

restlock = thread.allocate_lock()
locked = lockdec(restlock)

## RestResult class

class RestResult(LazyDict):

    def __init__(self, url="", name=""):
        LazyDict.__init__(self)
        self.url = url
        self.name = name
        self.data = None
        self.error = None
        self.status = None
        self.reason = ""

## RestClient class

class RestClient(object):

    """ Provide a REST client that works in sync mode. """

    def __init__(self, url, keyfile=None, certfile=None, port=None):
        if not url.endswith('/'): url += '/'
        try:
            u = urlparse(url)
            splitted = u[1].split(':')
            if len(splitted) == 2: host, port = splitted
            else:
                host = splitted[0]
                port = port or 9999
            path = u[2]
        except Exception, ex: raise
        self.host = host 
        try: self.ip = socket.gethostbyname(self.host)
        except Exception, ex: handle_exception()
        self.path = path
        self.port = port
        self.url = url
        self.keyfile = keyfile
        self.certfile = certfile
        self.callbacks = []

    def addcb(self, callback): 
        """ add a callback. """
        if not callback: return
        self.callbacks.append(callback)
        logging.debug('rest.client - added callback %s' % str(callback))
        return self

    def delcb(self, callback):
        """ delete callback. """
        try:
            del self.callbacks[callback]
            logging.debug('rest.client - deleted callback %s' % str(callback))
        except ValueError: pass
        
    def do(self, func, url, *args, **kwargs):
        """ perform a rest request. """
        result = RestResult(url)
        try:
            logging.info("rest.client - %s - calling %s" % (url, str(func)))
            res = func(url, {}, kwargs, self.keyfile, self.certfile, self.port)
            result.status = res.status
            result.reason = res.reason
            if result.status >= 400: result.error = result.status
            else: result.error = None
            if result.status == 200:
                r = res.read()
                result.data = loads(r)
            else: result.data = None
            logging.info("rest.client - %s - result: %s" % (url, str(result))) 
        except Exception, ex:
            result.error = str(ex)
            result.data = None
        for cb in self.callbacks:
            try:
                cb(self, result)
                logging.info('rest.client - %s - called callback %s' % (url, str(cb)))
            except Exception, ex:
                handle_exception()
        return result

    def post(self, *args, **kwargs):
        """ do a POST request. """
        return self.do(posturl, self.url, *args, **kwargs)

    def add(self, *args, **kwargs):
        """ add an REST item. """
        return self.do(posturl, self.url, *args, **kwargs)

    def delete(self, nr=None):
        """ delete a REST item. """
        if nr: return self.do(deleteurl, self.url + '/' + str(nr))
        else: return self.do(deleteurl, self.url)

    def get(self, nr=None):
        """ get a REST item. """
        if not nr: return self.do(geturl4, self.url)
        else: return self.do(geturl4, self.url + '/' + str(nr))

## RestClientAsync class

class RestClientAsync(RestClient, asynchat.async_chat):

    """ Async REST client. """

    def __init__(self, url, name=""):
        RestClient.__init__(self, url)
        asynchat.async_chat.__init__(self)
        self.set_terminator("\r\n\r\n")
        self.reading_headers = True
        self.error = None
        self.buffer = ''
        self.name = name or self.url
        self.headers = {}
        self.status = None

    def handle_error(self):
        """ take care of errors. """
        exctype, excvalue, tb = sys.exc_info()
        if exctype == socket.error:
            try:
                errno, errtxt = excvalue
                if errno in [11, 35, 9]:
                    logging.error("res.client - %s - %s %s" % (self.url, errno, errtxt))
                    return
            except ValueError: pass
            self.error = str(excvalue)
        else:
            logging.error("%s - %s" % (self.name, exceptionmsg()))
            self.error = exceptionmsg()
        self.buffer = ''
        result = RestResult(self.url, self.name)
        result.error = self.error
        result.data = None
        for cb in self.callbacks:
            try:
                cb(self, result)
                logging.info('rest.client - %s - called callback %s' % (url, str(cb)))
            except Exception, ex: handle_exception()
        self.close()

    def handle_expt(self):
        """ handle an exception. """
        handle_exception()

    def handle_connect(self):
        """ called after succesfull connect. """
        logging.info('rest.client - %s - connected %s' % (self.url, str(self)))
         
    def start(self):
        """ start the client loop. """
        assert(self.host)
        assert(int(self.port))
        try:
            logging.info('rest.client - %s - starting client' % self.url)
            self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
            self.connect((self.ip, int(self.port)))
        except socket.error, ex:
            self.error = str(ex)
            try:
                self.connect((self.ip, int(self.port)))
            except socket.error, ex: self.error = str(ex)
        except Exception, ex: self.error = str(ex)
        if self.error: self.warn("rest.client - %s - can't start %s" % (self.url, self.error))
        else: return True

    @locked
    def found_terminator(self):
        """ called when terminator is found. """
        logging.info('rest.client - %s - found terminator' % self.url)
        if self.reading_headers:
            self.reading_headers = False
            try:
                self.headers = self.buffer.split('\r\n')
                self.status = int(self.headers[0].split()[1])
            except (ValueError, IndexError):
                logging.warn("rest.client - %s - can't parse headers %s" % (self.url, self.headers))
                return
            self.set_terminator(None)
            self.buffer = ''
            logging.info('rest.client - %s - headers: %s' % (self.url, self.headers))

    def collect_incoming_data(self, data):
        """ aggregate seperate data chunks. """
        self.buffer = self.buffer + data

    def handle_close(self):
        """ called on connection close. """
        self.reading_headers = False
        self.handle_incoming()
        logging.info('rest.client - %s - closed' % self.url)
        self.close()
      
    def handle_incoming(self): 
        """ handle incoming data. """
        logging.info("rest.client - %s - incoming: %s" % (self.url, self.buffer))
        if not self.reading_headers:
            result = RestResult(self.url, self.name)
            if self.status >= 400:
                logging.warn('rest.client - %s - error status: %s' % (self.url, self.status))
                result.error = self.status
                result.data = None
            elif self.error:
                result.error = self.error
                result.data = None
            elif self.buffer == "":
                result.data = ""
                result.error = None
            else:
                try:
                    res = loads(self.buffer)
                    if not res:
                        self.buffer = ''
                        return
                    result.data = res
                    result.error = None
                except ValueError, ex:
                    logging.info("rest.client - %s - can't decode %s" % (self.url, self.buffer))
                    result.error = str(ex)
                except Exception, ex:
                    logging.error("rest.client - %s - %s" % (self.url, exceptionmsg()))
                    result.error = exceptionmsg()                
                    result.data = None
            for cb in self.callbacks:
                try:
                    cb(self, result)
                    logging.info('rest.client - %s - called callback %s' % (self.url, str(cb)))
                except Exception, ex: handle_exception()
            self.buffer = ''

    @locked
    def dorequest(self, method, path, postdata={}, headers={}):
        if postdata: postdata = urllib.urlencode(postdata)
        if headers:
            if not headers.has_key('Content-Length'): headers['Content-Length'] = len(postdata)
            headerstxt = ""
            for i,j in headers.iteritems(): headerstxt += "%s: %s\r\n" % (i.lower(), j)
        else: headerstxt = ""
        if method == 'POST': s = toenc("%s %s HTTP/1.0\r\n%s\r\n%s\r\n\r\n" % (method, path, headerstxt, postdata), 'ascii')
        else: s = toenc("%s %s HTTP/1.0\r\n\r\n" % (method, path), 'ascii')
        if self.start():
            logging.info('rest.client - %s - sending %s' % (self.url, s))
            self.push(s)

    def sendpost(self, postdata):
        headers = {'Content-Type': 'application/x-www-form-urlencoded', \
'Accept': 'text/plain; text/html', 'User-Agent': useragent()}
        self.dorequest('POST', self.path, postdata, headers)

    def sendget(self):
        """ send a GET request. """
        self.dorequest('GET', self.path)

    def post(self, *args, **kwargs):
        """ do a POST request. """
        self.sendpost(kwargs)

    def get(self):
        """ call GET request. """
        self.sendget()



# me2/plugs/fleet.py
#
#

""" 
    The fleet makes it possible to run multiple bots in one running instance.
    It is a list of bots.

"""

## me2 imports

from me2.lib.config import Config
from me2.lib.threads import start_new_thread
from me2.lib.fleet import getfleet, FleetBotAlreadyExists
from me2.lib.commands import cmnds
from me2.lib.examples import examples
from me2.utils.name import stripname

## basic imports

import os

## fleet-avail command

def handle_fleetavail(bot, ievent):
    """ show available fleet bots. """
    ievent.reply('available bots: ', getfleet().avail()) 

cmnds.add('fleet-avail', handle_fleetavail, 'OPER')
examples.add('fleet-avail', 'show available fleet bots', 'fleet-avail')

## fleet-connect command

def handle_fleetconnect(bot, ievent):

    """ connect a fleet bot to it's server. """
    try: botname = ievent.args[0]
    except IndexError:
        ievent.missing('<botname>')
        return
    try:
        fleet = getfleet()
        fleetbot = fleet.byname(botname)
        if fleetbot:
            start_new_thread(fleetbot.connect, ())
            ievent.reply('%s connect thread started' % botname)
        else:
            ievent.reply("can't connect %s .. trying enable" % botname)
            fleet.enable(bot, ievent)
    except Exception, ex:
        ievent.reply(str(ex))

cmnds.add('fleet-connect', handle_fleetconnect, 'OPER', threaded=True)
examples.add('fleet-connect', 'connect bot with <name> to irc server', 'fleet-connect test')

## fleet-disconnect command

def handle_fleetdisconnect(bot, ievent):
    """ disconnect a fleet bot from server. """
    try: botname = ievent.args[0]
    except IndexError:
        ievent.missing('<botname>')
        return
    ievent.reply('exiting %s' % botname)
    try:
        fleet = getfleet()
        if fleet.exit(botname): ievent.reply("%s bot stopped" % botname)
        else: ievent.reply("can't stop %s bot" % botname)
    except Exception, ex: ievent.reply("fleet - %s" % str(ex))

cmnds.add('fleet-disconnect', handle_fleetdisconnect, 'OPER', threaded=True)
examples.add('fleet-disconnect', 'fleet-disconnect <name> .. disconnect bot with <name> from irc server', 'fleet-disconnect test')

## fleet-list command

def handle_fleetlist(bot, ievent):
    """ fleet-list .. list bot names in fleet. """
    ievent.reply("fleet: ", getfleet().list())

cmnds.add('fleet-list', handle_fleetlist, ['USER', 'GUEST'])
examples.add('fleet-list', 'show current fleet list', 'fleet-list')

## fleet-del command

def handle_fleetdel(bot, ievent):
    """ delete bot from fleet. """
    try: name = ievent.args[0]
    except IndexError:
        ievent.missing('<name>')
        return
    try:
        if getfleet().delete(name): ievent.reply('%s deleted' % name)
        else: ievent.reply('%s delete failed' % name)
    except Exception, ex: ievent.reply(str(ex))

cmnds.add('fleet-del', handle_fleetdel, 'OPER', threaded=True)
examples.add('fleet-del', 'fleet-del <botname> .. delete bot from fleet list', 'fleet-del test')

## fleet-disable command

def fleet_disable(bot, ievent):
    """ disable a fleet bot. """
    if not ievent.rest:
        ievent.missing("list of fleet bots")
        return
    bots = ievent.rest.split()
    fleet = getfleet()
    for name in bots:
        bot = fleet.byname(name)
        if bot:
            bot.cfg['enable'] = 0
            bot.cfg.save()
            ievent.reply('disabled %s' % name)
            fleet.exit(name)
        else: ievent.reply("can't find %s bot in fleet" % name)

cmnds.add('fleet-disable', fleet_disable, 'OPER')
examples.add('fleet-disable', 'disable a fleet bot', 'fleet-disable local')

## fleet-enable command

def fleet_enable(bot, ievent):
    """ enable a fleet bot. """
    if not ievent.rest:
        ievent.missing("list of fleet bots")
        return
    bots = ievent.rest.split()
    fleet = getfleet()
    for name in bots:
        bot = fleet.byname(name)
        if bot:
            bot.cfg.load()
            bot.cfg['disable'] = 0
            bot.cfg.save()
            ievent.reply('enabled %s' % name)
            start_new_thread(bot.connect, ())
        elif name in fleet.avail():
            cfg = Config('fleet' + os.sep + stripname(name) + os.sep + 'config')
            cfg['disable'] = 0
            cfg.save()
            bot = fleet.makebot(cfg.type, cfg.name, cfg)
            ievent.reply('enabled and started %s bot' % name)
            start_new_thread(bot.connect, ())
        else: ievent.reply('no %s bot in fleet' % name)

cmnds.add('fleet-enable', fleet_enable, 'OPER', threaded=True)
examples.add('fleet-enable', 'enable a fleet bot', 'fleet-enable local')



# me2/plugs/gae.py
#
#

""" Google Application Engine related commands. """

## me2 imports

from me2.lib.commands import cmnds
from me2.lib.examples import examples

## commands

def handle_gaeflushcache(bot, ievent):
    """ flush the cache .. flush all with no arguments otherwise delete specific. """
    from google.appengine.api.memcache import flush_all, delete

    if not ievent.rest:
        flush_all()
    else:
        delete(ievent.rest)

    ievent.done()

cmnds.add('gae-flushcache', handle_gaeflushcache, 'OPER')
examples.add('gae-flushcache', 'flush the bots cache', 'gae-flushcache')

def handle_gaeadminstats(bot, ievent):
    """ show bot stats. """
    from google.appengine.api.memcache import get_stats
    ievent.reply("cache: %s" % str(get_stats()))

cmnds.add('gae-stats', handle_gaeadminstats, 'OPER')
examples.add('gae-stats', 'show bots stats', 'gae-stats')



# me2/plugs/remotecallbacks.py
#
#

""" dispatch remote events. """

## me2 imports

from me2.utils.lazydict import LazyDict
from me2.utils.generic import fromenc
from me2.utils.exception import handle_exception
from me2.lib.callbacks import callbacks, remote_callbacks, first_callbacks
from me2.lib.container import Container
from me2.lib.eventbase import EventBase
from me2.lib.errors import NoProperDigest
from me2.lib.commands import cmnds
from me2.lib.examples import examples
from me2.contrib.xmlstream import NodeBuilder, XMLescape, XMLunescape
from me2.contrib.simplejson import loads, dumps

## basic imports

import logging
import copy
import hmac
import hashlib
import cgi

## defines

cpy = copy.deepcopy

## callback

def remotecb(bot, event):
    """ dispatch an event. """
    try: container = Container().load(event.txt)
    except TypeError:
        handle_exception()
        logging.warn("remotecallbacks - not a remote event - %s " % event.userhost)
        return
    logging.debug('doing REMOTE callback')
    try:
        digest = hmac.new(str(container.hashkey), XMLunescape(container.payload), hashlib.sha512).hexdigest()
        logging.debug("remotecallbacks - digest is %s" % digest)
    except TypeError:
        handle_exception()
        logging.error("remotecallbacks - can't load payload - %s" % container.payload)
        return
    if container.digest == digest: e = EventBase().load(XMLunescape(container.payload))
    else: raise NoProperDigest()
    e.txt = XMLunescape(e.txt)
    e.nodispatch = True
    e.forwarded = True
    bot.doevent(e)
    event.status = "done"  
    return

remote_callbacks.add("MESSAGE", remotecb)
	


# me2/utils/mainloop.py
#
#

""" main loop used in me2 binairies. """

## me2 imports

from me2.lib.eventhandler import mainhandler
from me2.lib.exit import globalshutdown

## basic imports

import os
import time

## mainloop function

def mainloop():
    """ function to be used as mainloop. """
    while 1:
        try:
            time.sleep(1)
            mainhandler.handle_one()
        except KeyboardInterrupt: break
        except Exception, ex:
            handle_exception()
            globalshutdown()
            os._exit(1)
    globalshutdown()
    os._exit(0)



# me2.plugs.common/chatlog.py
#
#

"""
    log irc channels to [hour:min] <nick> txt format, only 
    logging to files is supported right now. 

"""

## me2 imports

from me2.lib.commands import cmnds
from me2.lib.callbacks import callbacks, remote_callbacks
from me2.lib.persistconfig import PersistConfig
from me2.utils.locking import lockdec
from me2.utils.timeutils import hourmin
from me2.lib.examples import examples
from me2.utils.exception import handle_exception
from me2.utils.lazydict import LazyDict
from me2.lib.datadir import getdatadir
from me2.utils.name import stripname

## basic imports

import time
import os
import logging
import thread
from os import path
from datetime import datetime

## locks

outlock = thread.allocate_lock()
outlocked = lockdec(outlock)

## defines

cfg = PersistConfig()
cfg.define('channels', [])
cfg.define('format', 'simple')
cfg.define('basepath', getdatadir())
cfg.define('nologprefix', '[nolog]')
cfg.define('nologmsg', '-= THIS MESSAGE NOT LOGGED =-')
cfg.define('backend', 'file')

logfiles = {}
backends = {}
stopped = False
db = None

# Formats are defined here. simple also provides default values if values
# are not supplied by the format, as well as format 'simple'. 
# Parameters that should be supplied:
#   * timestamp_format: format of timestamp in log files
#     * all strftime vars supported.
#   * filename: file name for log
#     * var channel : full channel ie. #dunkbot
#     * var channel_name : channel without '#' ie. dunkbot
#   * event_filename: 
#        if event_filename exists, then it will be used for
#        logging events (seperate from chat)
#     * var channel : full channel ie. #dunkbot
#     * var channel_name : channel without '#' ie. dunkbot
#   * separator: the separator between the timestamp and message

formats = {
    'simple': {
        'timestamp_format': '%Y-%m-%d %H:%M:%S',
        'basepath': None,
        'filename': 'chatlogs/%%(network)s/simple/%%(target)s.%Y%m%d.slog',
        'event_prefix': '',
        'event_filename': 'chatlogs/%%(network)s/simple/%%(channel_name)s.%Y%m%d.slog',
        'separator': ' | ',
    },
    'supy': {
        'timestamp_format': '%Y-%m-%dT%H:%M:%S',
        'filename': 'chatlogs/%%(network)s/supy/%%(target)s/%%(target)s.%Y-%m-%d.log',
        'event_prefix': '*** ',
        'event_filename': None,
        'separator': '  ',
    }
}

## functions

def format_opt(name):
    simple_format = formats['simple']
    format = formats.get(cfg.get('format'), 'simple')
    opt = format.get(name, simple_format.get(name, None))
    return opt

def init():
    global stopped
    stopped = False
    callbacks.add('ALL', chatlogcb, prechatlogcb)
    remote_callbacks.add('ALL', chatlogcb, prechatlogcb)
    return 1

def shutdown():
    global stopped
    stopped = True
    for file in logfiles.values():
        file.close()
    return 1

def timestr(dt):
    return dt.strftime(format_opt('timestamp_format'))

@outlocked
def write(m): 
    """
      m is a dict with the following properties:
      datetime
      type : (comment, nick, topic etc..)
      target : (#channel, bot etc..)
      txt : actual message
      network
    """
    backend_name = cfg.get('backend', 'file')
    backend = backends.get(backend_name, file_write)
    if m['txt'].startswith(cfg.get('nologprefix')): m['txt'] = cfg.get('nologmsg')
    backend(m)

def file_write(m):
    if stopped: return
    args = {
        'target': m.get('target'), 
        'network': m.get('network'), 
        'logdir': cfg.get('logdir')
    }
    #if args['target'][0] in "#":
    #    args['target'] = u"-" + args['target'][1:]
    args['channel_name'] = args['target'] 
    #else: args['channel_name'] = args['target']
    f = time.strftime(format_opt('filename')) % args
    if m['type'] != 'comment':
        event_filename = format_opt('event_filename')
        if event_filename: f = time.strftime(event_filename) % args
        m['txt'] = '%s%s'%(m['event_prefix'], m['txt'])
    else: m['txt'] = '<%s> %s'%(m['nick'], m['txt'])
    basepath = cfg.get('basepath')
    if basepath: f = path.join(basepath, f)
    dir = path.dirname(f)
    if not path.exists(dir):
         try: os.makedir(dir)
         except: pass
    timestamp = timestr(m['datetime'])
    line = '%(timestamp)s%(separator)s%(txt)s\n'%({
        'timestamp': timestamp, 
        'separator': format_opt('separator'),
        'txt': m['txt'],
    })
    try:
        if logfiles.has_key(f):
            logfiles[f].write(line)
            logfiles[f].flush()
        else:
            logging.warn('chatlog - opening %s for logging'%(f))
            logfiles[f] = open(f, 'a')
            logfiles[f].write(line)
            logfiles[f].flush()
    except Exception, ex: handle_exception()

backends['file'] = file_write

## log function

def formatevent(bot, ievent):
    m = {
        'datetime': datetime.now(),
        'separator': format_opt('separator'),
        'event_prefix': format_opt('event_prefix'),
        'network': bot.networkname,
        'nick': ievent.nick,
        'target': stripname(ievent.channel),
    }
    m = LazyDict(m)

    if ievent.cbtype == "OUTPUT":
        m.update({
            'type': 'output',
            'txt': ievent.txt
        })
    elif ievent.cmnd == 'PRIVMSG':
        if ievent.txt.startswith('\001ACTION'):
            m.update({
                'type': 'action',
                'txt': '* %s %s'%(m['nick'], ievent.txt[7:-1].strip()),
            })
        else:
            m.update({
                'type': 'comment',
                'txt': '%s'%(ievent.txt),
            })
    elif ievent.cmnd == 'NOTICE':
        m.update({
            'type': 'notice',
            'target': ievent.arguments[0],
            'txt': "-%s- %s"%(ievent.nick, ievent.txt)
        })
    elif ievent.cmnd == 'TOPIC':
        m.update({
            'type': 'topic',
            'txt': '%s changes topic to "%s"'%(ievent.nick, ievent.txt),
        })
    elif ievent.cmnd == 'MODE':
        margs = ' '.join(ievent.arguments[1:])
        m.update({
            'type': 'mode',
            'txt': '%s sets mode: %s'%(ievent.nick, margs),
        })
    elif ievent.cmnd == 'JOIN':
        m.update({
            'type': 'join',
            'txt': '%s (%s) has joined %s'%(ievent.nick, ievent.userhost, ievent.channel),
        })
    elif ievent.cmnd == 'KICK':
        m.update({
            'type': 'kick',
            'txt': '%s was kicked by %s (%s)'%(
                ievent.arguments[1], 
                ievent.nick, 
                ievent.txt
            ),
        })
    elif ievent.cmnd == 'PART':
        m.update({
            'type': 'part',
            'txt': '%s (%s) has left'%(ievent.nick, ievent.userhost),
        })
    elif ievent.cmnd in ('QUIT', 'NICK'):
        if not ievent.user or not ievent.user.data.channels:
            logging.debug("chatlog - can't find joined channels for %s" % ievent.userhost)
            return m
        cmd = ievent.cmnd
        nick = cmd == 'NICK' and ievent.txt or ievent.nick
        for c in event.user.channels:
            if [bot.name, c] in cfg.get('channels'):
                if True:
                    if cmd == 'NICK':
                        m['txt'] = '%s (%s) is now known as %s'%(
                            ievent.nick, ievent.userhost, ievent.txt
                        )
                    else:
                        m['txt'] = '%s (%s) has quit: %s'%(
                            ievent.nick, ievent.userhost, ievent.txt
                        )
                    m.update({
                        'type': ievent.cmnd.lower(),
                        'target': c,
                    })
                    write(m)
        return m
    elif ievent.cbtype == 'MESSAGE':
            m.update({
                'type': 'comment',
                'txt': ievent.txt.strip(),
            })
    elif ievent.cbtype == 'PRESENCE':
            if ievent.type == 'unavailable':
                m.update({
                    'type': 'part',
                    'txt': "%s left"%ievent.nick
                })
            else:
                m.update({
                    'type': 'join',
                    'txt': "%s joined"%ievent.nick
                })
    elif ievent.cbtype == 'BLIP_SUBMITTED':
            m.update({
                'type': 'blip',
                'txt': ievent.txt.strip(),
            })
    elif ievent.cbtype == 'DISPATCH':
            m.update({
                'type': 'dipatch',
                'txt': ievent.txt,
            })

    return m

def log(bot, event):
    m = formatevent(bot, event)
    if m.get('txt'):
        write(m)

## chatlog precondition

def prechatlogcb(bot, ievent):
    """Check if event should be logged.  QUIT and NICK are not channel
    specific, so we will check each channel in log()."""
    if bot.isgae: return False
    if not cfg.channels:
        return False
    if not ievent.msg and [bot.name, ievent.channel] in cfg.get('channels'):
        return True
    if ievent.cmnd in ('QUIT', 'NICK'):
        return True
    if ievent.cmnd == 'NOTICE':
        if [bot.name, ievent.arguments[0]] in cfg.get('channels'):
            return True

## chatlog callbacks

def chatlogcb(bot, ievent):
    log(bot, ievent)

## chatlog-on command

def handle_chatlogon(bot, ievent):
    """ enable chatlog. """
    chan = ievent.channel
    if [bot.name, chan] not in cfg.get('channels'):
        cfg['channels'].append([bot.name, chan])
        cfg.save()
        ievent.reply('chatlog enabled on (%s,%s)' % (bot.name, chan))
    else: ievent.reply('chatlog already enabled on (%s,%s)' % (bot.name, chan))

cmnds.add('chatlog-on', handle_chatlogon, 'OPER')
examples.add('chatlog-on', 'enable chatlog on <channel> or the channel \
the commands is given in', '1) chatlog-on 2) chatlog-on #dunkbots')

## chatlog-off command

def handle_chatlogoff(bot, ievent):
    """ disable chatlog. """
    try:
        cfg['channels'].remove([bot.name, ievent.channel])
        cfg.save()
    except ValueError:
        ievent.reply('chatlog is not enabled in (%s,%s)' % (bot.name, ievent.channel))
        return
    ievent.reply('chatlog disabled on (%s,%s)' % (bot.name, ievent.channel))

cmnds.add('chatlog-off', handle_chatlogoff, 'OPER')
examples.add('chatlog-off', 'disable chatlog on <channel> or the channel the commands is given in', '1) chatlog-off 2) chatlog-off #dunkbots')



# me2/plugs/gatekeeper.py
#
#

""" gatekeeper commands. """

## me2 imports

from me2.lib.commands import cmnds
from me2.lib.examples import examples

## gatekeeper-allow command

def handle_gatekeeperallow(bot, event):
    """ allow user on bot. """
    if not event.rest:
        event.missing("<userhost>")
        return
    bot.gatekeeper.allow(event.rest)
    event.done()

cmnds.add('gatekeeper-allow', handle_gatekeeperallow, 'OPER')
examples.add('gatekeeper-allow', 'add JID of remote bot that we allow to receice events from', 'gatekeeper-allow me2@appspot.com')

## gatekeeper-deny command

def handle_gatekeeperdeny(bot, event):
    """ deny user on bot. """
    if not event.rest:
        event.missing("<userhost>")
        return
    bot.gatekeeper.deny(event.rest)
    event.done()

cmnds.add('gatekeeper-deny', handle_gatekeeperdeny, 'OPER')
examples.add('gatekeeper-deny', 'remove JID of remote bot', 'gatekeeper-deny evilfscker@pissof.com')



# me2/socklib/xmpp/message.py
#
#

""" jabber message definition .. types can be normal, chat, groupchat, 
    headline or  error
"""

## me2 imports

from me2.utils.exception import handle_exception
from me2.utils.trace import whichmodule
from me2.utils.generic import toenc, fromenc, jabberstrip
from me2.utils.locking import lockdec
from me2.lib.eventbase import EventBase
from me2.lib.errors import BotNotSetInEvent
from me2.lib.gozerevent import GozerEvent

## xmpp import

from me2.contrib.xmlstream import NodeBuilder, XMLescape, XMLunescape

## basic imports

import types
import time
import thread
import logging
import re

## locks

replylock = thread.allocate_lock()
replylocked = lockdec(replylock)

## classes

class Message(GozerEvent):

    """ jabber message object. """

    def __init__(self, nodedict={}):
        self.element = "message"
        self.jabber = True
        self.cmnd = "MESSAGE"
        self.cbtype = "MESSAGE"
        self.bottype = "xmpp"
        self.type = "normal"
        GozerEvent.__init__(self, nodedict)
  
    def __copy__(self):
        return Message(self)

    def __deepcopy__(self, bla):
        m = Message()
        m.copyin(self)
        return m

    def parse(self, bot=None):
        """ set ircevent compat attributes. """
        self.bot = bot
        self.jidchange = False
        self.cmnd = 'MESSAGE'
        try: self.resource = self.fromm.split('/')[1]
        except IndexError: pass
        self.channel = self['fromm'].split('/')[0]
        self.origchannel = self.channel
        self.nick = self.resource
        self.jid = self.fromm
        self.ruserhost = self.jid
        self.userhost = self.jid
        self.stripped = self.jid.split('/')[0]
        self.printto = self.channel
        for node in self.subelements:
            try:
                self.txt = node.body.data
                break
            except (AttributeError, ValueError):
                continue
        if self.txt: self.usercmnd = self.txt.split()[0]
        else: self.usercmnd = ""
        self.origtxt = self.txt
        self.time = time.time()
        if self.type == 'groupchat':
            self.groupchat = True
            self.auth = self.userhost
        else:
            self.groupchat = False
            self.auth = self.stripped
        self.msg = not self.groupchat
        self.makeargs()

    def errorHandler(self):
        """ dispatch errors to their handlers. """
        try:
            code = self.get('error').code
        except Exception, ex:
            handle_exception()
        try:
            method = getattr(self, "handle_%s" % code)
            if method:
                logging.error('sxmpp.core - dispatching error to handler %s' % str(method))
                method(self)
        except AttributeError, ex: logging.error('sxmpp.core - unhandled error %s' % code)
        except: handle_exception()

    def normalize(self, what):
        return self.bot.normalize(what)



# me2.plugs.wave/clone.py
#
#

""" clone the wave after x blips. """

## me2 imports

from me2.lib.commands import cmnds
from me2.lib.callbacks import callbacks
from me2.lib.gae.wave.waves import Wave
from me2.lib.plugins import plugs

## basic imports

import logging

## callbacks

def clonepre(bot, event):
    if event.bottype == "wave":
        return True

def clonecallback(bot, event):
    wave = event.chan
    if wave.data.threshold != -1 and (wave.data.seenblips > wave.data.threshold):
        wave.data.threshold = -1
        newwave = wave.clone(bot, event, event.title)
        plugs.load('me2.plugs.common.hubbub')
        feeds = plugs['me2.plugs.common.hubbub'].watcher.clone(bot.name, bot.type, newwave.data.waveid, event.waveid)
        event.reply("this wave is continued to %s with the following feeds: %s" % (newwave.data.url, feeds))

#callbacks.add("BLIP_SUBMITTED", clonecallback, clonepre)
#callbacks.add('OUTPUT', clonecallback, clonepre)



#!/usr/bin/python2.4
#
# Copyright (C) 2009 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for the wavelet module."""


import unittest

import blip
import element
import ops
import wavelet

import simplejson

ROBOT_NAME = 'robot@appspot.com'

TEST_WAVELET_DATA = {
    'creator': ROBOT_NAME,
    'creationTime': 100,
    'lastModifiedTime': 101,
    'participants': [ROBOT_NAME],
    'participantsRoles': {ROBOT_NAME: wavelet.Participants.ROLE_FULL},
    'rootBlipId': 'blip-1',
    'title': 'Title',
    'waveId': 'test.com!w+g3h3im',
    'waveletId': 'test.com!root+conv',
    'tags': ['tag1', 'tag2'],
}

TEST_BLIP_DATA = {
    'blipId': TEST_WAVELET_DATA['rootBlipId'],
    'childBlipIds': [],
    'content': '\ntesting',
    'contributors': [TEST_WAVELET_DATA['creator'], 'robot@google.com'],
    'creator': TEST_WAVELET_DATA['creator'],
    'lastModifiedTime': TEST_WAVELET_DATA['lastModifiedTime'],
    'parentBlipId': None,
    'waveId': TEST_WAVELET_DATA['waveId'],
    'elements': {},
    'waveletId': TEST_WAVELET_DATA['waveletId'],
}


class TestWavelet(unittest.TestCase):
  """Tests the wavelet class."""

  def setUp(self):
    self.operation_queue = ops.OperationQueue()
    self.all_blips = {}
    self.blip = blip.Blip(TEST_BLIP_DATA,
                          self.all_blips,
                          self.operation_queue)
    self.all_blips[self.blip.blip_id] = self.blip
    self.wavelet = wavelet.Wavelet(TEST_WAVELET_DATA,
                                   self.all_blips,
                                   None,
                                   self.operation_queue)
    self.wavelet.robot_address = ROBOT_NAME

  def testWaveletProperties(self):
    w = self.wavelet
    self.assertEquals(TEST_WAVELET_DATA['creator'], w.creator)
    self.assertEquals(TEST_WAVELET_DATA['creationTime'], w.creation_time)
    self.assertEquals(TEST_WAVELET_DATA['lastModifiedTime'],
                      w.last_modified_time)
    self.assertEquals(len(TEST_WAVELET_DATA['participants']),
                      len(w.participants))
    self.assertTrue(TEST_WAVELET_DATA['participants'][0] in w.participants)
    self.assertEquals(TEST_WAVELET_DATA['rootBlipId'], w.root_blip.blip_id)
    self.assertEquals(TEST_WAVELET_DATA['title'], w.title)
    self.assertEquals(TEST_WAVELET_DATA['waveId'], w.wave_id)
    self.assertEquals(TEST_WAVELET_DATA['waveletId'], w.wavelet_id)
    self.assertEquals('test.com', w.domain)

  def testWaveletMethods(self):
    w = self.wavelet
    reply = w.reply()
    self.assertEquals(2, len(w.blips))
    w.delete(reply)
    self.assertEquals(1, len(w.blips))
    self.assertEquals(0, len(w.data_documents))
    self.wavelet.data_documents['key'] = 'value'
    self.assert_('key' in w.data_documents)
    self.assertEquals(1, len(w.data_documents))
    for key in w.data_documents:
      self.assertEquals(key, 'key')
    self.assertEquals(1, len(w.data_documents.keys()))
    self.wavelet.data_documents['key'] = None
    self.assertEquals(0, len(w.data_documents))
    num_participants = len(w.participants)
    w.proxy_for('proxy').reply()
    self.assertEquals(2, len(w.blips))
    # check that the new proxy for participant was added
    self.assertEquals(num_participants + 1, len(w.participants))
    w._robot_address = ROBOT_NAME.replace('@', '+proxy@')
    w.proxy_for('proxy').reply()
    self.assertEquals(num_participants + 1, len(w.participants))
    self.assertEquals(3, len(w.blips))

  def testSetTitle(self):
    self.blip._content = '\nOld title\n\nContent'
    self.wavelet.title = 'New title \xd0\xb0\xd0\xb1\xd0\xb2'
    self.assertEquals(1, len(self.operation_queue))
    self.assertEquals('wavelet.setTitle',
                      self.operation_queue.serialize()[1]['method'])
    self.assertEquals(u'\nNew title \u0430\u0431\u0432\n\nContent',
                      self.blip._content)

  def testSetTitleAdjustRootBlipWithOneLineProperly(self):
    self.blip._content = '\nOld title'
    self.wavelet.title = 'New title'
    self.assertEquals(1, len(self.operation_queue))
    self.assertEquals('wavelet.setTitle',
                      self.operation_queue.serialize()[1]['method'])
    self.assertEquals('\nNew title\n', self.blip._content)

  def testSetTitleAdjustEmptyRootBlipProperly(self):
    self.blip._content = '\n'
    self.wavelet.title = 'New title'
    self.assertEquals(1, len(self.operation_queue))
    self.assertEquals('wavelet.setTitle',
                      self.operation_queue.serialize()[1]['method'])
    self.assertEquals('\nNew title\n', self.blip._content)

  def testTags(self):
    w = self.wavelet
    self.assertEquals(2, len(w.tags))
    w.tags.append('tag3')
    self.assertEquals(3, len(w.tags))
    w.tags.append('tag3')
    self.assertEquals(3, len(w.tags))
    w.tags.remove('tag1')
    self.assertEquals(2, len(w.tags))
    self.assertEquals('tag2', w.tags[0])

  def testParticipantRoles(self):
    w = self.wavelet
    self.assertEquals(wavelet.Participants.ROLE_FULL,
                      w.participants.get_role(ROBOT_NAME))
    w.participants.set_role(ROBOT_NAME, wavelet.Participants.ROLE_READ_ONLY)
    self.assertEquals(wavelet.Participants.ROLE_READ_ONLY,
                      w.participants.get_role(ROBOT_NAME))

  def testSerialize(self):
    self.blip.append(element.Gadget('http://test.com', {'a': 3}))
    self.wavelet.title = 'A wavelet title'
    self.blip.append(element.Image(url='http://www.google.com/logos/clickortreat1.gif',
                              width=320, height=118))
    self.blip.append(element.Attachment(caption='fake', data='fake data'))
    self.blip.append(element.Line(line_type='li', indent='2'))
    self.blip.append('bulleted!')
    self.blip.append(element.Installer(
        'http://wave-skynet.appspot.com/public/extensions/areyouin/manifest.xml'))
    self.wavelet.proxy_for('proxy').reply().append('hi from douwe')
    inlineBlip = self.blip.insert_inline_blip(5)
    inlineBlip.append('hello again!')

    serialized = self.wavelet.serialize()
    serialized = simplejson.dumps(serialized)
    self.assertTrue(serialized.find('test.com') > 0)

if __name__ == '__main__':
  unittest.main()



# me2/gae/tasks.py
#
#

""" appengine tasks related classes and functions. """

## me2 imports

from me2.utils.exception import handle_exception

## google imports

from google.appengine.api.labs.taskqueue import Task, Queue

## simplejson imports

from me2.contrib.simplejson import dumps

## basic imports

import uuid

## Event Classes

class BotEvent(Task):
    pass

## defines

queues = []
for i in range(9):
    queues.append(Queue("queue" + str(i)))

## start_botevent function

def start_botevent(bot, event, speed=5):
    """ start a new botevent task. """
    try:
        event.botevent = True
        if event.usercmnd[0] == "!": e = event.usercmnd[1:]
        else: e = event.usercmnd
        name = e + "-" + str(uuid.uuid4())
        payload = dumps({ 'bot': bot.tojson(),
                          'event': event.tojson()
                        })
        be = BotEvent(name=name, payload=payload, url="/tasks/botevent")
        try: queues[int(speed)].add(be)
        except TypeError: queues[speed].add(be)
    except Exception, ex: 
        handle_exception()
